import torch
import torch.nn as nn
from torch.nn.modules.loss import _Loss


class Hamiltonian(_Loss):
    def __init__(self, layer, reg_cof=1e-4):
        super(Hamiltonian, self).__init__()
        self.layer = layer
        self.reg_cof = 0

    def forward(self, x, p):
        y = self.layer(x)
        # l2 = cal_l2_norm(self.layer)

        # print(y.shape, p.shape)
        H = torch.sum(y * p)

        # H = H - self.reg_cof * l2
        return H


class CrossEntropyWithWeightPenlty(_Loss):
    def __init__(self, module, DEVICE, reg_cof=1e-4):
        super(CrossEntropyWithWeightPenlty, self).__init__()

        self.reg_cof = reg_cof
        self.criterion = nn.CrossEntropyLoss().to(DEVICE)
        self.module = module
        # print(modules, 'dwadaQ!')

    def __call__(self, pred, label):
        cross_loss = self.criterion(pred, label)
        weight_loss = 0
        # for module in self.module:
        #    print(module)
        #    weight_loss = weight_loss + cal_l2_norm(module)

        weight_loss = cal_l2_norm(self.module)

        loss = cross_loss + self.reg_cof * weight_loss
        return loss


def cal_l2_norm(layer: torch.nn.Module):
    loss = 0.0
    for name, param in layer.named_parameters():
        if name == "weight":
            loss = (
                loss
                + 0.5
                * torch.norm(
                    param,
                )
                ** 2
            )

    return loss



#############################################################################################
#  Copied and modified from https://github.com/bearpaw/pytorch-classification/blob/master/utils/misc.py  #
#############################################################################################

"""Some helper functions for PyTorch, including:
- get_mean_and_std: calculate the mean and std value of dataset.
- msr_init: net parameter initialization.
- progress_bar: progress bar mimic xlua.progress.
"""

import errno
import os

import numpy as np
import torch
import torch.nn as nn
import torch.nn.init as init

__all__ = ["get_mean_and_std", "init_params", "mkdir_p", "AverageMeter"]


def get_mean_and_std(dataset):
    """Compute the mean and std value of dataset."""
    dataloader = torch.utils.data.DataLoader(
        dataset, batch_size=1, shuffle=True, num_workers=2
    )

    mean = torch.zeros(3)
    std = torch.zeros(3)
    print("==> Computing mean and std..")
    for inputs, targets in dataloader:
        for i in range(3):
            mean[i] += inputs[:, i, :, :].mean()
            std[i] += inputs[:, i, :, :].std()
    mean.div_(len(dataset))
    std.div_(len(dataset))
    return mean, std


class RunningMeanStd:
    def __init__(self, dim=1):
        self._mean = np.zeros(dim)
        self._count = 0
        self._M = np.zeros(dim)
        self.dim = dim

    def update(self, x):
        """
        :param x: [n, d]
        :return:  None
        """
        if isinstance(x, list):
            x = np.array(x)

        avg_a = self._mean
        avg_b = np.mean(x, axis=0)

        count_a = self._count
        count_b = x.shape[0]

        delta = avg_b - avg_a
        m_a = self._M
        m_b = np.var(x, axis=0) * count_b
        M2 = (
            m_a
            + m_b
            + np.power(delta, 2) * count_a * count_b / (count_a + count_b)
        )

        self._mean += delta * count_b / (count_a + count_b)
        self._M = M2
        self._count += count_b

    @property
    def mean(self):
        return self._mean

    @property
    def std(self):
        if self._count == 1:
            return np.ones(self.dim)
        return np.sqrt(self._M / (self._count - 1))


def get_mean_and_std_modified(dataset):
    # Compute the mean and std value online
    """Compute the mean and std value of dataset."""
    dataloader = torch.utils.data.DataLoader(
        dataset, batch_size=1, shuffle=True, num_workers=2
    )
    rms = RunningMeanStd(dim=3)

    print("==> Computing mean and std..")
    for inputs, targets in dataloader:
        inputs = inputs.detach().cpu().numpy()
        inputs = inputs.transpose((0, 2, 3, 1)).reshape(-1, 3)
        rms.update(inputs)
    return rms.mean, rms.std


def init_params(net):
    """Init layer parameters."""
    for m in net.modules():
        if isinstance(m, nn.Conv2d):
            init.kaiming_normal(m.weight, mode="fan_out")
            if m.bias:
                init.constant(m.bias, 0)
        elif isinstance(m, nn.BatchNorm2d):
            init.constant(m.weight, 1)
            init.constant(m.bias, 0)
        elif isinstance(m, nn.Linear):
            init.normal(m.weight, std=1e-3)
            if m.bias:
                init.constant(m.bias, 0)


def mkdir_p(path):
    """make dir if not exist"""
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


class AverageMeter:
    """Computes and stores the average and current value
    Imported from https://github.com/pytorch/examples/blob/master/imagenet/main.py#L247-L262
    """

    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count



# (c) 2024 LAiSR-SK
# This code is licensed under the MIT license (see LICENSE.md).
"""Some utility functions"""

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F


def one_hot_tensor(y_batch_tensor, num_classes, device):
    y_tensor = torch.cuda.FloatTensor(
        y_batch_tensor.size(0), num_classes
    ).fill_(0)
    y_tensor[np.arange(len(y_batch_tensor)), y_batch_tensor] = 1.0
    return y_tensor


def label_smoothing(y_batch_tensor, num_classes, delta):
    y_batch_smooth = (
        1 - delta - delta / (num_classes - 1)
    ) * y_batch_tensor + delta / (num_classes - 1)
    return y_batch_smooth


def str2bool(v):
    return v.lower() in ("yes", "true", "t", "1")


class softCrossEntropy(nn.Module):
    def __init__(self, reduce=True):
        super(softCrossEntropy, self).__init__()
        self.reduce = reduce
        return

    def forward(self, inputs, targets):
        """
        :param inputs: predictions
        :param targets: target labels in vector form
        :return: loss
        """
        log_likelihood = -F.log_softmax(inputs, dim=1)
        sample_num, class_num = targets.shape
        if self.reduce:
            loss = torch.sum(torch.mul(log_likelihood, targets)) / sample_num
        else:
            loss = torch.sum(torch.mul(log_likelihood, targets), 1)

        return loss


class CWLoss(nn.Module):
    def __init__(self, num_classes, margin=50, reduce=True):
        super(CWLoss, self).__init__()
        self.num_classes = num_classes
        self.margin = margin
        self.reduce = reduce
        return

    def forward(self, logits, targets):
        """
        :param inputs: predictions
        :param targets: target labels
        :return: loss
        """
        onehot_targets = one_hot_tensor(
            targets, self.num_classes, targets.device
        )

        self_loss = torch.sum(onehot_targets * logits, dim=1)
        other_loss = torch.max(
            (1 - onehot_targets) * logits - onehot_targets * 1000, dim=1
        )[0]

        loss = -torch.sum(torch.clamp(self_loss - other_loss + self.margin, 0))

        if self.reduce:
            sample_num = onehot_targets.shape[0]
            loss = loss / sample_num

        return loss



# (c) 2024 LAiSR-SK
# This code is licensed under the MIT license (see LICENSE.md).
import copy
import os
import time
from collections import namedtuple
from typing import Dict, NamedTuple

import numpy as np
import torch
from autoattack import AutoAttack
from torch import Tensor, nn, optim
from torch.nn import Module
from torch.optim import Optimizer
from torch.utils.data import DataLoader

from airtk.attack import create_attack
from airtk.helper_functions import (
    cw_whitebox,
    eval_clean,
    get_model,
    load_data,
    mim_whitebox,
    robust_eval,
)
from airtk.loss import accuracy, standard_loss


class VaTraining:
    """Class which provides a simple interface for the VA training method."""

    def __init__(
        self,
        dataset: str,
        model: str,
        batch_size: int = 128,
        epochs: int = 200,
        weight_decay: float = 2e-4,
        lr: float = 0.1,
        momentum: float = 0.9,
        seed: int = 1,
        log_interval: int = 1,
        model_dir: str = "data/model",
        save_freq: int = 1,
        swa_warmup: int = 50,
        swa: bool = True,
        lr_schedule: str = "decay",
    ) -> None:
        """Creates a new `VA` instance.

        :param dataset: the dataset to be used as a str. Should be either
            "cifar10" or "cifar100".
        :param model: the model to train. Should be one of
            "res18", "res34", "res50", "res101", "res152", or "wideres34".
        :param batch_size: the size of batches for training and testing.
            Defaults to 128.
        :param epochs: the number of epochs to train for.
            Defaults to 200.
        :param weight_decay: a weight decay constant for regularization.
            Defaults to 2e-4.
        :param lr: the initial learning rate. Defaults to 0.1.
        :param momentum: Nesterov momentum constant. Defaults to 0.9.
        :param seed: the random seed to use for training. Defaults to 1.
        :param log_interval: the number of batches to log per. Defaults to 1.
        :param model_dir: the the directory to output models to.
            Defaults to "data/model".
        :param swa_warmup: the number of epochs to apply
            stochastic weight averaging to. Defaults to 50.
        :param swa: whether or not to apply stochastic weight averaging.
            Defaults to True.
        :param lr_schedule: the learning rate schedule. Should be one of
            "decay", "schedules", "cosine", "cyclic-10", or "cyclic-5".
            Defaults to "decay".
        """
        ArgsPrototype = namedtuple(
            "ArgsPrototype",
            [
                "batch_size",
                "test_batch_size",
                "epochs",
                "weight_decay",
                "lr",
                "momentum",
                "no_cuda",
                "seed",
                "log_interval",
                "model_dir",
                "save_freq",
                "swa_warmup",
                "swa",
                "lr_schedule",
            ],
        )
        self._args = ArgsPrototype(
            batch_size,
            batch_size,
            epochs,
            weight_decay,
            lr,
            momentum,
            False,
            seed,
            log_interval,
            model_dir,
            save_freq,
            swa_warmup,
            swa,
            lr_schedule,
        )
        self._dataset = dataset
        self._model = model

    def __call__(self) -> None:
        """Trains the model defined by __init__."""
        main(self._dataset, self._model, 0, self._args)


def train(
    args: NamedTuple,
    model: Module,
    device: torch.device,
    train_loader: DataLoader,
    optimizer: Optimizer,
    epoch: int,
    attacks: Dict[int, str],
    ds_name: str,
):
    """
    Trains one epoch by calculating the loss for each batch based on the given
    set of attacks and updating the model

    :param args: set of arguments including learning rate, log interval, etc
    :param model: model to train
    :param device: the current device
    :param train_loader: data loader containing the training dataset
    :param optimizer: optimizer to train
    :param epoch: current epoch of training
    :param attacks: dictionary of attack types based on class
    :param ds_name: name of dataset
    """

    model.train()  # Set the model to training mode

    # For each batch of images in the train loader
    for batch_idx, sample in enumerate(train_loader):
        # Extract the image and classification details
        if ds_name == "cifar100":
            data, target, coarse = sample
        else:
            data, target = sample

        # Set the data/target to the current device
        data, target = (
            data.to(device, non_blocking=True),
            target.to(device, non_blocking=True),
        )

        # Determine the 'class' of the batch by looking at the most common
        # class
        if ds_name == "cifar100":
            avg_cls = int(torch.mode(coarse)[0])
        else:
            avg_cls = int(torch.mode(target)[0])

        # Determine the attack based on the most common class
        attack = attacks[avg_cls]

        optimizer.zero_grad()  # zero optimizer gradients

        # Calculate loss using the specified attack
        loss, batch_metrics = va_loss(
            model, data, target, optimizer, attack, device, ds_name
        )

        # Update model parameters
        loss.backward()
        optimizer.step()

        # Print training progress
        if batch_idx % args.log_interval == 0:
            print(
                f"Train Epoch: {epoch} [{batch_idx * len(data)}/"
                + f"{len(train_loader.dataset)} ("
                + f"{100.0 * batch_idx / len(train_loader):.0f}%)]\tLoss: "
                + f"{loss.item():.6f}"
            )


def va_loss(
    model: Module,
    x: Tensor,
    y: Tensor,
    optimizer: Optimizer,
    att: str,
    device: torch.device,
    ds: str,
    keep_clean: bool = False,
):
    """
    Adversarial training loss function for later various atts epoch, which
    simply calculates the loss based on the given attack.

    :param model: trained image classifier
    :param x: batch of clean images
    :param y: labels associated with images in x
    :param optimizer: optimizer to be updated
    :param att: attack to be used
    :param device: current device
    :param ds: name of dataset
    :param keep_clean: True if clean examples are kept, False default

    :return loss to the classifier
    :return metrics of the batch
    """

    # Set the criterion to cross-entropy loss
    criterion = nn.CrossEntropyLoss()

    # Create the adversary for AutoAttack
    adversary = AutoAttack(
        model, norm="Linf", eps=0.031, version="custom", verbose=False
    )

    # Generate adversarial example based on specified attack
    if att == "none":
        return standard_loss(model, x, y, optimizer)
    elif att == "linf-pgd-40":
        attack = create_attack(model, criterion, "linf-pgd", 0.03, 40, 0.01)
        x_adv, _ = attack.perturb(x, y)
    elif att == "linf-pgd-20":
        attack = create_attack(model, criterion, "linf-pgd", 0.03, 20, 0.01)
        x_adv, _ = attack.perturb(x, y)
    elif att == "linf-pgd-7":
        attack = create_attack(model, criterion, "linf-pgd", 0.03, 7, 0.01)
        x_adv, _ = attack.perturb(x, y)
    elif att == "linf-pgd-3":
        attack = create_attack(model, criterion, "linf-pgd", 0.03, 3, 0.01)
        x_adv, _ = attack.perturb(x, y)
    elif att == "fgsm-40":
        attack = create_attack(model, criterion, "fgsm", 0.03, 40, 0.01)
        x_adv, _ = attack.perturb(x, y)
    elif att == "fgsm-20":
        attack = create_attack(model, criterion, "fgsm", 0.03, 20, 0.01)
        x_adv, _ = attack.perturb(x, y)
    elif att == "fgsm-7":
        attack = create_attack(model, criterion, "fgsm", 0.03, 7, 0.01)
        x_adv, _ = attack.perturb(x, y)
    elif att == "cw":
        x_adv = cw_whitebox(model, x, y, device, ds)
    elif att == "mim":
        x_adv = mim_whitebox(model, x, y, device)
    elif att == "l2-pgd-40":
        attack = create_attack(model, criterion, "l2-pgd", 0.03, 40, 0.01)
        x_adv, _ = attack.perturb(x, y)
    elif att == "l2-pgd-20":
        attack = create_attack(model, criterion, "l2-pgd", 0.03, 20, 0.01)
        x_adv, _ = attack.perturb(x, y)
    elif att == "l2-pgd-7":
        attack = create_attack(model, criterion, "l2-pgd", 0.03, 7, 0.01)
        x_adv, _ = attack.perturb(x, y)
    elif att == "autoattack":
        aa_attack = AutoAttack(
            model, norm="Linf", eps=0.031, version="standard", verbose=False
        )
        x_adv = aa_attack.run_standard_evaluation(x, y)
    elif att == "apgd-ce":
        adversary.attacks_to_run = ["apgd-ce"]
        x_adv = adversary.run_standard_evaluation(x, y)
    elif att == "apgd-dlr":
        adversary.attacks_to_run = ["apgd-dlr"]
        x_adv = adversary.run_standard_evaluation(x, y)
    elif att == "apgd-t":
        adversary.attacks_to_run = ["apgd-t"]
        x_adv = adversary.run_standard_evaluation(x, y)
    elif att == "fab":
        adversary.attacks_to_run = ["fab"]
        x_adv = adversary.run_standard_evaluation(x, y)
    elif att == "fab-t":
        adversary.attacks_to_run = ["fab-t"]
        x_adv = adversary.run_standard_evaluation(x, y)
    elif att == "square":
        adversary.attacks_to_run = ["square"]
        x_adv = adversary.run_standard_evaluation(x, y)
    else:
        print(att)
        raise NotImplementedError

    optimizer.zero_grad()  # zero out the gradients

    # Correctly format x_adv and y_adv
    if keep_clean:
        x_adv = torch.cat((x, x_adv), dim=0)
        y_adv = torch.cat((y, y), dim=0)
    else:
        y_adv = y

    # Calculates the adversarial output and CE loss
    out = model(x_adv)
    criterion = nn.CrossEntropyLoss()
    loss = criterion(out, y_adv)

    preds = out.detach()  # correctly format adversarial output

    # Calculate batch metrics to be returned
    batch_metrics = {"loss": loss.item()}
    if keep_clean:
        preds_clean, preds_adv = preds[: len(x)], preds[len(x) :]
        batch_metrics.update(
            {
                "clean_acc": accuracy(y, preds_clean),
                "adversarial_acc": accuracy(y, preds_adv),
            }
        )
    else:
        batch_metrics.update({"adversarial_acc": accuracy(y, preds)})

    return loss, batch_metrics


def adjust_learning_rate(
    optimizer: Optimizer, epoch: int, args: NamedTuple
) -> None:
    """
    Sets the learning rate of the optimizer based on the current epoch

    :param optimizer: optimizer with learning rate being set
    :param epoch: current epoch
    :param args: program arguments
    """
    lr = args.lr
    if args.lr_schedule == "decay":
        if epoch >= 50:
            lr = args.lr * 0.1
        if epoch >= 75:
            lr = args.lr * 0.01
    elif args.lr_schedule == "scheduled":
        if epoch >= 24:
            lr = args.lr * 0.1
        if epoch >= 26:
            lr = args.lr
        if epoch >= 64:
            lr = args.lr * 0.1
        if epoch >= 66:
            lr = args.lr
        if epoch >= 104:
            lr = args.lr * 0.1
        if epoch >= 106:
            lr = args.lr
        if epoch >= 137:
            lr = args.lr * 0.1
        if epoch >= 139:
            lr = args.lr * 0.01
    elif args.lr_schedule == "cosine":
        lr = 0.2
        lr = lr * 0.5 * (1 + np.cos((epoch - 1) / args.epochs * np.pi))
    elif args.lr_schedule == "cyclic-10":
        if epoch % 10 < 2 or epoch % 10 == 9:
            lr = args.lr
        elif epoch % 10 < 4 or epoch % 10 > 6:
            lr = args.lr * 0.1
        elif epoch % 10 == 5:
            lr = args.lr * 0.001
        else:
            lr = args.lr * 0.01
    elif args.lr_schedule == "cyclic-5":
        if epoch % 5 == 0:
            lr = args.lr
        elif epoch % 5 == 1 or epoch % 5 == 4:
            lr = args.lr * 0.1
        else:
            lr = args.lr * 0.01
    for param_group in optimizer.param_groups:
        param_group["lr"] = lr


def epochs_define_attacks(epoch: int, dataset: str) -> Dict[int, str]:
    """
    Defines a dictionary of attacks based on the current epoch.

    :param epoch: current epoch
    :param dataset: name of dataset

    :return attacks: dictionary of attacks
    """

    attacks = {}
    if dataset == "cifar10":
        if epoch <= 5:
            for i in range(4):
                attacks[i] = "l2-pgd-7"
            for i in range(4, 7):
                attacks[i] = "l2-pgd-40"
            for i in range(7, 10):
                attacks[i] = "l2-pgd-20"
        elif epoch <= 10:
            for i in range(4):
                attacks[i] = "l2-pgd-20"
            for i in range(4, 7):
                attacks[i] = "l2-pgd-7"
            for i in range(7, 10):
                attacks[i] = "l2-pgd-40"
        elif epoch <= 15:
            for i in range(4):
                attacks[i] = "l2-pgd-40"
            for i in range(4, 7):
                attacks[i] = "l2-pgd-20"
            for i in range(7, 10):
                attacks[i] = "l2-pgd-7"
        elif epoch <= 25:
            for i in range(10):
                attacks[i] = "linf-pgd-3"
        elif epoch <= 30:
            for i in range(3):
                attacks[i] = "linf-pgd-7"
            for i in range(3, 6):
                attacks[i] = "linf-pgd-20"
            for i in range(6, 8):
                attacks[i] = "linf-pgd-40"
            for i in range(8, 10):
                attacks[i] = "mim"
        elif epoch <= 35:
            for i in range(3):
                attacks[i] = "linf-pgd-20"
            for i in range(3, 6):
                attacks[i] = "linf-pgd-7"
            for i in range(6, 8):
                attacks[i] = "mim"
            for i in range(8, 10):
                attacks[i] = "linf-pgd-40"
        elif epoch <= 40:
            for i in range(3):
                attacks[i] = "linf-pgd-40"
            for i in range(3, 6):
                attacks[i] = "mim"
            for i in range(6, 8):
                attacks[i] = "linf-pgd-7"
            for i in range(8, 10):
                attacks[i] = "linf-pgd-20"
        elif epoch <= 45:
            for i in range(3):
                attacks[i] = "mim"
            for i in range(3, 6):
                attacks[i] = "linf-pgd-40"
            for i in range(6, 8):
                attacks[i] = "linf-pgd-20"
            for i in range(8, 10):
                attacks[i] = "linf-pgd-7"
        elif epoch <= 53:
            for i in range(2):
                attacks[i] = "apgd-t"
            for i in range(2, 5):
                attacks[i] = "cw"
            for i in range(5, 8):
                attacks[i] = "apgd-dlr"
            for i in range(8, 10):
                attacks[i] = "apgd-ce"
        elif epoch <= 61:
            for i in range(2):
                attacks[i] = "apgd-ce"
            for i in range(2, 5):
                attacks[i] = "apgd-t"
            for i in range(5, 8):
                attacks[i] = "cw"
            for i in range(8, 10):
                attacks[i] = "apgd-dlr"
        elif epoch <= 68:
            for i in range(2):
                attacks[i] = "apgd-dlr"
            for i in range(2, 5):
                attacks[i] = "apgd-ce"
            for i in range(5, 8):
                attacks[i] = "apgd-t"
            for i in range(8, 10):
                attacks[i] = "cw"
        elif epoch <= 75:
            for i in range(2):
                attacks[i] = "cw"
            for i in range(2, 5):
                attacks[i] = "apgd-dlr"
            for i in range(5, 8):
                attacks[i] = "apgd-ce"
            for i in range(8, 10):
                attacks[i] = "apgd-t"
        else:
            for i in range(10):
                attacks[i] = "autoattack"
    elif dataset == "cifar100":
        if epoch <= 25:
            for i in range(20):
                attacks[i] = "linf-pgd-3"
        elif epoch <= 45:
            for i in range(10):
                attacks[i] = "linf-pgd-7"
            for i in range(10, 20):
                attacks[i] = "mim"
        elif epoch <= 65:
            for i in range(10):
                attacks[i] = "mim"
            for i in range(10, 20):
                attacks[i] = "linf-pgd-7"
        elif epoch <= 85:
            for i in range(10):
                attacks[i] = "linf-pgd-20"
            for i in range(10, 20):
                attacks[i] = "linf-pgd-40"
        elif epoch <= 105:
            for i in range(10):
                attacks[i] = "linf-pgd-40"
            for i in range(10, 20):
                attacks[i] = "linf-pgd-20"
        elif epoch <= 140:
            for i in range(20):
                attacks[i] = "cw"
        else:
            for i in range(20):
                attacks[i] = "autoattack"
    else:
        raise NotImplementedError

    return attacks


def main_warmup(
    ds_name: str, mod_name: str, epochs: int, args: NamedTuple
) -> None:
    """
    Main training method, which establishes the model/dataset before
    conducting training and clean testing for each epoch. Then reports final
    training time and conducts robustness tests on the final model.

    :param ds_name: dataset to use for training
    :param mod_name: model to use for training
    :param epochs: number of epochs to train for
    """
    device = "cuda" if torch.cuda.is_available else "cpu"
    model_dir = args.model_dir

    # Create file to print training progress
    filename = f"log/clean-{ds_name}-{mod_name}-output.txt"

    kwargs = (
        {"num_workers": 1, "pin_memory": True}
        if torch.cuda.is_available()
        else {}
    )
    with open(filename, "a") as f:
        # Initialize the model based on the specified architecture
        start_tot = time.time()  # start recording training time
        model = get_model(mod_name, ds_name, device)

        # Set up the optimizer with the arguments
        optimizer = optim.SGD(
            model.parameters(),
            lr=args.lr,
            momentum=args.momentum,
            weight_decay=args.weight_decay,
        )

        # Set up the dataloaders
        train_loader, test_loader = load_data(
            ds_name, args, kwargs, coarse=True
        )

        # Begin model training for the specified number of epochs
        for epoch in range(1, epochs + 1):
            # Adjust the SGD learning rate based on the epoch
            adjust_learning_rate(optimizer, epoch, args)

            start = time.time()  # start recording training time
            train(args, model, device, train_loader, optimizer, epoch, ds_name)
            end = time.time()  # stop recording training time
            epoch_time = end - start

            # Evaluation on natural and adversarial examples
            print(
                "================================================================"
            )
            f.write("Epoch " + str(epoch) + "\n")
            print("Time for Training: " + str(epoch_time))
            f.write("Time for Training: " + str(epoch_time) + "\n")
            eval_clean(model, device, train_loader, "train", ds_name, f)
            eval_clean(model, device, test_loader, "test", ds_name, f)
            robust_eval(model, device, test_loader, ds_name, f)
            print(
                "================================================================"
            )

            # Save the model (if designated)
            if epoch % args.save_freq == 0:
                torch.save(
                    model.state_dict(),
                    os.path.join(
                        model_dir,
                        f"model-warmup-{ds_name}-{mod_name}-epoch{epoch}.pt",
                    ),
                )
                torch.save(
                    optimizer.state_dict(),
                    os.path.join(
                        model_dir,
                        f"opt-warmup-{ds_name}-{mod_name}-epoch{epoch}.tar",
                    ),
                )

        # Report full training time
        end_tot = time.time()
        total_time = end_tot - start_tot
        f.write("Total training time: " + str(total_time) + "\n")

    f.close()  # close output file


def main(ds_name, mod_name, clean_epochs, args) -> None:
    """
    Main training method, which establishes the model/dataset before
    conducting training and clean testing for each epoch. Then reports final
    training time and conducts robustness tests on the final model.

    :param ds_name: dataset (cifar10 or cifar100)
    :param mod_name: training model architecture
    :param clean_pochs: number of epochs in warmup round, default 0
    """

    model_dir = args.model_dir
    if not os.path.exists(model_dir):
        os.makedirs(model_dir)
    use_cuda = not args.no_cuda and torch.cuda.is_available()
    torch.manual_seed(args.seed)
    device = torch.device("cuda" if use_cuda else "cpu")
    kwargs = {"num_workers": 1, "pin_memory": True} if use_cuda else {}

    # Create file to print training progress
    filename = f"log/va-epochs-{ds_name}-{mod_name}-output.txt"
    with open(filename, "a") as f:
        # Initialize the model based on the specified architecture
        start_tot = time.time()  # start recording training time
        base_model = get_model(mod_name, ds_name, device)

        # Clean training/warmup round, if specified
        if clean_epochs == 0:
            model = copy.deepcopy(base_model).to(device)
        else:
            main_warmup(ds_name, mod_name, clean_epochs)
            base_model.load_state_dict(
                torch.load(
                    os.path.join(
                        model_dir,
                        f"model-warmup-{ds_name}-{mod_name}-epoch{clean_epochs}.pt",
                    )
                )
            )
            model = copy.deepcopy(base_model).to(device)

        # Set up the optimizer with the arguments
        optimizer = optim.SGD(
            model.parameters(),
            lr=args.lr,
            momentum=args.momentum,
            weight_decay=args.weight_decay,
        )

        # Set up the dataloaders
        train_loader, test_loader = load_data(
            ds_name, args, kwargs, coarse=True
        )

        # Save the untrained model and optimizer
        torch.save(
            model.state_dict(),
            os.path.join(
                model_dir,
                f"model-va-epochs-{ds_name}-{mod_name}-start.pt",
            ),
        )
        torch.save(
            optimizer.state_dict(),
            os.path.join(
                model_dir,
                f"opt-va-epochs-{ds_name}-{mod_name}-start.tar",
            ),
        )

        # Begin model training
        for epoch in range(1, args.epochs + 1):
            # Adjust the SGD learning rate based on the epoch
            adjust_learning_rate(optimizer, epoch, args)

            # Define the dictionary of attack types for the current epoch
            attacks = epochs_define_attacks(epoch, ds_name)

            # Calculate the time elapsed
            start = time.time()  # start recording training time
            train(
                args,
                model,
                device,
                train_loader,
                optimizer,
                epoch,
                attacks,
                ds_name,
            )
            end = time.time()  # stop recording training time
            epoch_time = end - start

            # Evaluation on natural and adversarial examples
            f.write("Epoch " + str(epoch) + "\n")
            f.write("Current Attack Dictionary: ")
            f.write(str(attacks) + "\n")
            f.write("Time for Training: " + str(epoch_time) + "\n")
            eval_clean(model, device, train_loader, "train", ds_name, f)
            eval_clean(model, device, test_loader, "test", ds_name, f)
            robust_eval(model, device, test_loader, ds_name, f)

            # Save the model (if designated)
            if epoch % args.save_freq == 0:
                torch.save(
                    model.state_dict(),
                    os.path.join(
                        model_dir,
                        f"model-va-epochs-{ds_name}-{mod_name}-epoch{epoch}.pt",
                    ),
                )
                torch.save(
                    optimizer.state_dict(),
                    os.path.join(
                        model_dir,
                        f"opt-va-epochs-{ds_name}-{mod_name}-epoch{epoch}.tar",
                    ),
                )

        # Report full training time
        end_tot = time.time()
        total_time = end_tot - start_tot
        print("Total training time: " + str(total_time))
        f.write("Total training time: " + str(total_time) + "\n")

        # Load the final trained model, make a deepcopy, and evaluate against
        # FGSM and PGD attacks
        trained_model = get_model(mod_name, ds_name, device)

        trained_model.load_state_dict(
            torch.load(
                os.path.join(
                    model_dir,
                    f"model-va-epochs-{ds_name}-{mod_name}-epoch{epoch}.pt",
                ),
            )
        )
        model_copy = copy.deepcopy(trained_model)
        robust_eval(model_copy, device, test_loader, ds_name, f)



# (c) 2024 LAiSR-SK

# This code is licensed under the MIT license (see LICENSE.md).
import argparse
import os
import sys
import time

import numpy as np
import torch
import torch.optim as optim
from airtk.model import WideResNet
from torch import nn
from torch.optim import lr_scheduler
from torch.utils.data import DataLoader, Dataset
from torchvision import datasets, transforms

parser = argparse.ArgumentParser()
parser.add_argument(
    "--seed_torch", type=int, default=np.random.randint(429496729)
)
parser.add_argument(
    "--seed_numpy", type=int, default=np.random.randint(429496729)
)

parser.add_argument("--gpus", type=str, default="0")

parser.add_argument("--epochs", type=int, default=200)

parser.add_argument("--batch_size", type=int, default=128)

parser.add_argument("--checkpoint", type=str, default=".log/custat_output")
parser.add_argument(
    "--loss_type", type=str, default="xent", choices=["xent", "mix"]
)

## Inner maximization

parser.add_argument("--num_steps", type=int, default=10)

parser.add_argument("--alpha", type=float, default=2)

parser.add_argument("--epsilon", type=float, default=8)

parser.add_argument("--eta", type=float, default=5e-3)

parser.add_argument("--c", type=float, default=10)

parser.add_argument("--kappa", type=float, default=10)

parser.add_argument("--epsilon_max", type=float, default=8)

## Optimizer

parser.add_argument("--lr", type=float, default=0.1)

parser.add_argument("--momentum", type=float, default=0.9)

parser.add_argument("--weight_decay", type=float, default=5e-4)
args = parser.parse_args()


np.random.seed(args.seed_numpy)
torch.manual_seed(args.seed_torch)

os.environ["CUDA_VISIBLE_DEVICES"] = args.gpus


def func(epoch):
    if epoch < 79:
        return 1.0

    elif epoch < 139:
        return 0.1

    elif epoch < 179:
        return 0.1**2

    else:
        return 0.1**3


######################################

# Custom Dataloader
######################################


class Cat_dataloader(Dataset):
    def __init__(self, data, is_train=True, transform=None):
        self.data = datasets.__dict__[data.upper()](
            "./data", train=is_train, download=True, transform=transform
        )

    def __len__(self):
        return len(self.data)

    def __getitem__(self, i):
        image = self.data[i][0]

        label = self.data[i][1]

        index = i

        return image, label, index


#########################################

# Label Smoothing using Dirichlet dist.
#########################################


def label_smoothing(targets, epsilon, c, num_classes):
    onehot = torch.eye(num_classes, device="cuda")[targets].cuda()

    dirich = torch.from_numpy(
        np.random.dirichlet(np.ones(num_classes), targets.size(0))
    ).cuda()
    sr = (
        (torch.ones(targets.size(0)).cuda() * (c * epsilon))
        .unsqueeze(1)
        .repeat(1, num_classes)
    )

    ones = torch.ones_like(sr)

    y_tilde = (ones - sr) * onehot + sr * dirich

    return y_tilde


#########################################

# Loss functions
#########################################


class Loss_func:
    def __init__(self, num_classes, loss_type="xent", kappa=0):
        self.loss_type = loss_type
        self.num_classes = num_classes

        self.loss_type = loss_type

        self.kappa = kappa

    def xent(self, logits, targets):
        probs = torch.softmax(logits, dim=1)

        loss = -torch.sum(targets * torch.log(probs)) / probs.size(0)
        return loss

    def mix(self, logits, targets, org_targets):
        probs = torch.softmax(logits, dim=1)

        batch = probs.size(0)

        class_index = (
            torch.arange(self.num_classes)[None, :].repeat(batch, 1).cuda()
        )

        false_probs = torch.topk(
            probs[class_index != org_targets[:, None]].view(
                batch, self.num_classes - 1
            ),
            k=1,
        ).values

        gt_probs = probs[class_index == org_targets[:, None]].unsqueeze(1)

        cw_loss = torch.max(
            (false_probs - gt_probs).view(-1),
            self.kappa * torch.ones(batch).cuda(),
        )
        loss = torch.sum(
            torch.sum(-targets * torch.log(probs), dim=1) + cw_loss
        ) / probs.size(0)
        return loss

    def __call__(self, logits, targets, org_targets):
        if self.loss_type == "xent":
            return self.xent(logits, targets)

        elif self.loss_type == "mix":
            return self.mix(logits, targets, org_targets)


#########################################

# Inner Maximization
#########################################


def inner_maximization(
    model, loss_func, inputs, targets, org_targets, epsilons, alpha, num_steps
):
    # Properly format the epsilon values

    epsilons = epsilons[:, None, None, None].repeat(
        1, inputs.size(1), inputs.size(2), inputs.size(3)
    )

    for _ in range(num_steps):
        x = inputs.requires_grad_()  # get gradients of x

        logits = model(x)  # determine the output of x
        loss = loss_func(
            logits, targets, org_targets
        )  # calculate loss between output, smoothed targets, + unsmoothed

        loss.backward()

        grads = x.grad.data

        x = (
            x.data.detach() + alpha * torch.sign(grads).detach()
        )  # perturb x using the gradient

        x = torch.min(
            torch.max(x, inputs - epsilons), inputs + epsilons
        )  # constrain x using epsilon

        x = torch.clamp(x, min=0, max=1)  # clamp x between 0 and 1

    return x


def train(
    epoch,
    model,
    dataloader,
    optimizer,
    num_classes,
    loss_func,
    epsilons,
    alpha,
    num_steps,
    c,
    eta,
    epsilon_max,
):
    model.train()

    for idx, samples in enumerate(dataloader):
        start_time = time.time()  # start timing
        inputs, targets, indices = (
            samples  # get data, targets, and locations for each image
        )
        inputs = inputs.cuda()
        targets = targets.cuda()
        indices = indices.cuda()

        smoothed_targets = label_smoothing(
            targets, epsilons[indices], c, num_classes
        )  # get smoothed labels

        epsilons[indices] += (
            eta  # add eta to the epsilon at the index of each image
        )

        x = inner_maximization(
            model,
            loss_func,
            inputs,
            smoothed_targets,
            targets,
            epsilons[indices],
            alpha,
            num_steps,
        )  # get the new x

        logits = model(x)  # calc output of new x

        t_or_f = torch.argmax(torch.softmax(logits, dim=1), dim=1).eq(
            targets
        )  # determine if output=targets

        false_indices = indices[
            torch.where(t_or_f == False)[0]
        ]  # find outputs that don't equal targets

        epsilons[false_indices] -= eta  # subtract eta from wrong epsilons

        # constrain all epsilons by the min of the current epsilon value and the max epsilon

        epsilons[indices] = torch.min(
            epsilons[indices],
            (torch.ones(inputs.size(0)) * epsilon_max).cuda(),
        )

        smoothed_targets = label_smoothing(
            targets, epsilons[indices], c, num_classes
        )  # smooth targets with new e's
        loss = loss_func(
            logits, smoothed_targets, targets
        )  # calculate the loss

        optimizer.zero_grad()

        loss.backward()

        optimizer.step()

        # calculate robust accuracy in training
        num_correct = (
            torch.argmax(torch.softmax(logits, dim=1), dim=1)
            .eq(targets)
            .sum()
            .item()
        )

        train_rob_acc = 100 * (num_correct / inputs.size(0))

        end_time = time.time()

        elapsed_time = end_time - start_time

        if idx % 100 == 0:
            print(
                "Epoch %d [%d/%d] | loss: %.4f | rob acc: %.4f | elapsed time: %.4f"
                % (
                    epoch,
                    idx + 1,
                    len(dataloader),
                    loss.item(),
                    train_rob_acc,
                    elapsed_time,
                )
            )


def evaluation(epoch, model, dataloader, loss_func, epsilon, alpha, num_steps):
    model.eval()

    counter = 0

    total_corr_nat = 0

    total_corr_rob = 0

    xent = nn.CrossEntropyLoss()

    for idx, samples in enumerate(dataloader):
        inputs, targets, _ = samples
        inputs = inputs.cuda()
        targets = targets.cuda()

        total_corr_nat += (
            torch.argmax(torch.softmax(model(inputs), dim=1), dim=1)
            .eq(targets)
            .sum()
            .item()
        )

        noise = (
            torch.FloatTensor(inputs.size()).uniform_(-epsilon, epsilon).cuda()
        )

        x = torch.clamp(inputs + noise, min=0, max=1)

        for _ in range(num_steps):
            x.requires_grad_()

            logits = model(x)

            loss = xent(logits, targets)

            loss.backward()

            grads = x.grad.data

            x = x.data.detach() + alpha * torch.sign(grads).detach()

            x = torch.min(torch.max(x, inputs - epsilon), inputs + epsilon)

            x = torch.clamp(x, min=0, max=1)

        total_corr_rob += (
            torch.argmax(torch.softmax(model(x), dim=1), dim=1)
            .eq(targets)
            .sum()
            .item()
        )

        counter += inputs.size(0)

        sys.stdout.write(
            f"\r [Eval] [{idx * len(inputs)}/{len(dataloader.dataset)} ({100.0 * idx / len(dataloader):.0f}%)]\tacc nat: {100 * (total_corr_nat / counter):.5f}[%]\tacc rob: {100 * (total_corr_rob / counter):.5f}[%]"
        )

    avg_nat = 100 * (total_corr_nat / len(dataloader.dataset))

    avg_rob = 100 * (total_corr_rob / len(dataloader.dataset))
    print()

    return avg_nat, avg_rob


def main_custat(ds_name):
    out_dir = os.path.join(args.checkpoint, ds_name, args.loss_type)

    os.makedirs(out_dir, exist_ok=True)

    if ds_name == "cifar10":
        num_classes = 10

    elif ds_name == "cifar100":
        num_classes = 100

    else:
        raise NotImplementedError

    train_transforms = transforms.Compose(
        [
            transforms.RandomCrop(32, padding=4),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
        ]
    )

    train_dataset = Cat_dataloader(
        ds_name, is_train=True, transform=train_transforms
    )

    test_dataset = Cat_dataloader(
        ds_name, is_train=False, transform=transforms.ToTensor()
    )

    train_dataloader = DataLoader(
        train_dataset,
        batch_size=args.batch_size,
        shuffle=True,
        drop_last=False,
        num_workers=os.cpu_count(),
    )

    test_dataloader = DataLoader(
        test_dataset,
        batch_size=args.batch_size,
        shuffle=False,
        drop_last=False,
        num_workers=os.cpu_count(),
    )

    model = nn.DataParallel(
        WideResNet(
            depth=34, num_classes=num_classes, widen_factor=10, dropRate=0.0
        ).cuda()
    )

    optimizer = optim.SGD(
        model.parameters(),
        lr=args.lr,
        momentum=args.momentum,
        weight_decay=args.weight_decay,
    )

    adjust_lr = lr_scheduler.LambdaLR(optimizer, lr_lambda=func)

    loss_func = Loss_func(
        num_classes=num_classes, loss_type=args.loss_type, kappa=args.kappa
    )

    # logger_test = Logger(os.path.join(out_dir, 'log_results.txt'))

    # logger_test.set_names(['Epoch', 'Natural Test Acc', 'PGD20 Acc'])

    best_acc = 0

    epsilons = torch.zeros(len(train_dataloader.dataset)).cuda()

    for epoch in range(args.epochs):
        train(
            epoch,
            model,
            train_dataloader,
            optimizer,
            num_classes,
            loss_func,
            epsilons,
            args.alpha / 255,
            args.num_steps,
            args.c,
            args.eta,
            args.epsilon_max / 255,
        )

        avg_nat, avg_rob = evaluation(
            epoch,
            model,
            test_dataloader,
            loss_func,
            args.epsilon / 255,
            args.alpha / 255,
            20,
        )

        # logger_test.append([epoch + 1, avg_nat, avg_rob])

        print("Epoch: " + str(epoch))

        print("avg_nat: " + str(avg_nat))

        print("avg_rob: " + str(avg_nat))  # added print statements vs logger

        if avg_rob > best_acc:
            best_acc = avg_rob

            best_checkpoint = {
                "epoch": epoch + 1,
                "state_dict": model.state_dict(),
                "test_nat_acc": avg_nat,
                "test_rob_acc": avg_rob,
                "optimizer": optimizer.state_dict(),
            }

            torch.save(
                best_checkpoint, os.path.join(out_dir, "bestpoint.pth.tar")
            )

        std_checkpoint = {
            "epoch": epoch + 1,
            "state_dict": model.state_dict(),
            "test_nat_acc": avg_nat,
            "test_rob_acc": avg_rob,
            "optimizer": optimizer.state_dict(),
        }

        torch.save(
            std_checkpoint, os.path.join(out_dir, "std_checkpoint.pth.tar")
        )

        adjust_lr.step()


if __name__ == "__main__":
    main_custat("cifar10")



import torch
import torch.nn as nn

from ...attacks import PGD

from .advtrainer import AdvTrainer


class AT(AdvTrainer):
    r"""
    Adversarial training in 'Towards Deep Learning Models Resistant to Adversarial Attacks'
    [https://arxiv.org/abs/1706.06083]

    Attributes:
        self.rmodel : rmodel.
        self.device : device where rmodel is.
        self.optimizer : optimizer.
        self.scheduler : scheduler (Automatically updated).
        self.curr_epoch : current epoch starts from 1 (Automatically updated).
        self.curr_iter : current iters starts from 1 (Automatically updated).

    Arguments:
        rmodel (nn.Module): rmodel to train.
        eps (float): strength of the attack or maximum perturbation.
        alpha (float): step size.
        steps (int): number of steps.
        random_start (bool): using random initialization of delta.
    """

    def __init__(self, rmodel, eps, alpha, steps, random_start=True):
        super().__init__(rmodel)
        self.atk = PGD(rmodel, eps, alpha, steps, random_start)

    def calculate_cost(self, train_data, reduction="mean"):
        r"""
        Overridden.
        """
        images, labels = train_data
        images = images.to(self.device)
        labels = labels.to(self.device)

        adv_images = self.atk(images, labels)
        logits_adv = self.rmodel(adv_images)

        cost = nn.CrossEntropyLoss(reduction="none")(logits_adv, labels)
        self.add_record_item("CALoss", cost.mean().item())

        return cost.mean() if reduction == "mean" else cost



# from easydict import EasyDict
import os
import sys

import numpy as np
import torch


def add_path(path):
    if path not in sys.path:
        print(f"Adding {path}")

        sys.path.append(path)


abs_current_path = os.path.realpath("./")

root_path = os.path.join("/", *abs_current_path.split(os.path.sep)[:-3])

lib_dir = os.path.join(root_path, "lib")

add_path(lib_dir)


from airtk.defense.yopo.training_config import (
    IPGDAttackMethodMaker,
    PieceWiseConstantLrSchedulerMaker,
    SGDOptimizerMaker,
    TrainingConfigBase,
)


class TrainingConfig(TrainingConfigBase):
    lib_dir = lib_dir

    def __init__(
        self,
        eps: float,
        num_epochs: int,
        val_interval: int,
        weight_decay: float,
        inner_iters: int,
        K: int,
        sigma: float,
        nb_iters: int,
        gamma: float,
        momentum: float,
    ) -> None:
        super().__init__()
        self.eps = eps
        self.num_epochs = num_epochs
        self.val_interval = val_interval
        self.weight_decay = weight_decay
        self.inner_iters = inner_iters
        self.K = K
        self.sigma = sigma
        self.nb_iters = nb_iters
        self.gamma = gamma
        self.momentum = momentum

    num_epochs = 36

    val_interval = 2

    weight_decay = 5e-4

    inner_iters = 3

    K = 5

    sigma = 2 / 255.0

    eps = 8 / 255.0

    nb_iters = 20

    def create_optimizer(self):
        return SGDOptimizerMaker(
            lr=1e-1 * 4 / self.K,
            momentum=self.momentum,
            weight_decay=self.weight_decay,
        )

    def create_lr_scheduler(self):
        return PieceWiseConstantLrSchedulerMaker(
            milestones=[30, 34, 36], gamma=self.gamma
        )

    create_loss_function = torch.nn.CrossEntropyLoss

    # create_attack_method = \

    #    IPGDAttackMethodMaker(eps = 8/255.0, sigma = 2/255.0, nb_iters = 10, norm = np.inf,

    #                          mean = torch.tensor(np.array([0]).astype(np.float32)[np.newaxis, :, np.newaxis, np.newaxis]),

    #                          std = torch.tensor(np.array([1]).astype(np.float32)[np.newaxis, :, np.newaxis, np.newaxis]))

    create_attack_method = None

    def create_evaluation_attack_method(self):
        return IPGDAttackMethodMaker(
            eps=self.eps,
            sigma=self.sigma,
            nb_iters=self.nb_iters,
            norm=np.inf,
            mean=torch.tensor(
                np.array([0]).astype(np.float32)[
                    np.newaxis, :, np.newaxis, np.newaxis
                ]
            ),
            std=torch.tensor(
                np.array([1]).astype(np.float32)[
                    np.newaxis, :, np.newaxis, np.newaxis
                ]
            ),
        )


# About data

# C.inp_chn = 1

# C.num_class = 10


# parser = argparse.ArgumentParser()

# parser.add_argument(
#
# "--resume",

# default=None,

# type=str,

# metavar="PATH",

# help="path to latest checkpoint (default: none)",
# )
# parser.add_argument(

# "-b",

# "--batch_size",

# default=256,

# type=int,

# metavar="N",

# help="mini-batch size",
# )

# parser.add_argument("-d", type=int, default=0, help="Which gpu to use")
# parser.add_argument(

# "-adv_coef",

# default=1.0,

# type=float,

# help="Specify the weight for adversarial loss",
# )
# parser.add_argument(

# "--auto-continue",

# default=False,
# action="store_true",

# help="Continue from the latest checkpoint",
# )
# args = parser.parse_args()


# if __name__ == "__main__":
# pass



import torch
import torch.nn as nn

from ..attack import Attack


class BIM(Attack):
    r"""
    BIM or iterative-FGSM in the paper 'Adversarial Examples in the Physical World'
    [https://arxiv.org/abs/1607.02533]

    Distance Measure : Linf

    Arguments:
        model (nn.Module): model to attack.
        eps (float): maximum perturbation. (Default: 8/255)
        alpha (float): step size. (Default: 2/255)
        steps (int): number of steps. (Default: 10)

    .. note:: If steps set to 0, steps will be automatically decided following the paper.

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.BIM(model, eps=8/255, alpha=2/255, steps=10)
        >>> adv_images = attack(images, labels)
    """

    def __init__(self, model, eps=8 / 255, alpha=2 / 255, steps=10):
        super().__init__("BIM", model)
        self.eps = eps
        self.alpha = alpha
        if steps == 0:
            self.steps = int(min(eps * 255 + 4, 1.25 * eps * 255))
        else:
            self.steps = steps
        self.supported_mode = ["default", "targeted"]

    def forward(self, images, labels):
        r"""
        Overridden.
        """

        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        if self.targeted:
            target_labels = self.get_target_label(images, labels)

        loss = nn.CrossEntropyLoss()

        ori_images = images.clone().detach()

        for _ in range(self.steps):
            images.requires_grad = True
            outputs = self.get_logits(images)

            # Calculate loss
            if self.targeted:
                cost = -loss(outputs, target_labels)
            else:
                cost = loss(outputs, labels)

            # Update adversarial images
            grad = torch.autograd.grad(
                cost, images, retain_graph=False, create_graph=False
            )[0]

            adv_images = images + self.alpha * grad.sign()
            a = torch.clamp(ori_images - self.eps, min=0)
            b = (adv_images >= a).float() * adv_images + (
                adv_images < a
            ).float() * a  # nopep8
            c = (b > ori_images + self.eps).float() * (ori_images + self.eps) + (
                b <= ori_images + self.eps
            ).float() * b  # nopep8
            images = torch.clamp(c, max=1).detach()

        return images



from collections import OrderedDict

import torch
from tqdm import tqdm

from airtk.defense.yopo.utils_misc import torch_accuracy


class FastGradientLayerOneTrainer:
    def __init__(
        self,
        Hamiltonian_func,
        param_optimizer,
        inner_steps=2,
        sigma=0.008,
        eps=0.03,
    ):
        self.inner_steps = inner_steps
        self.sigma = sigma
        self.eps = eps
        self.Hamiltonian_func = Hamiltonian_func
        self.param_optimizer = param_optimizer

    def step(self, inp, p, eta):
        """
        Perform Iterative Sign Gradient on eta

        ret: inp + eta
        """

        p = p.detach()

        for i in range(self.inner_steps):
            tmp_inp = inp + eta
            tmp_inp = torch.clamp(tmp_inp, 0, 1)
            H = self.Hamiltonian_func(tmp_inp, p)

            eta_grad_sign = torch.autograd.grad(
                H, eta, only_inputs=True, retain_graph=False
            )[0].sign()

            eta = eta - eta_grad_sign * self.sigma

            eta = torch.clamp(eta, -1.0 * self.eps, self.eps)
            eta = torch.clamp(inp + eta, 0.0, 1.0) - inp
            eta = eta.detach()
            eta.requires_grad_()
            eta.retain_grad()

        # self.param_optimizer.zero_grad()

        yofo_inp = eta + inp
        yofo_inp = torch.clamp(yofo_inp, 0, 1)

        loss = -1.0 * self.Hamiltonian_func(yofo_inp, p)

        loss.backward()
        # self.param_optimizer.step()
        # self.param_optimizer.zero_grad()

        return yofo_inp, eta


def train_one_epoch(
    net,
    batch_generator,
    optimizer,
    criterion,
    LayerOneTrainner,
    K,
    eps: float,
    DEVICE=torch.device("cuda:0"),
    descrip_str="Training",
):
    """

    :param attack_freq:  Frequencies of training with adversarial examples. -1 indicates natural training
    :param AttackMethod: the attack method, None represents natural training
    :return:  None    #(clean_acc, adv_acc)
    """
    net.train()
    pbar = tqdm(batch_generator)
    yofoacc = -1
    cleanacc = -1
    cleanloss = -1
    pbar.set_description(descrip_str)
    for i, (data, label) in enumerate(pbar):
        data = data.to(DEVICE)
        label = label.to(DEVICE)

        eta = torch.FloatTensor(*data.shape).uniform_(-eps, eps)
        eta = eta.to(label.device)
        eta.requires_grad_()

        optimizer.zero_grad()
        LayerOneTrainner.param_optimizer.zero_grad()

        for j in range(K):
            # optimizer.zero_grad()

            pbar_dic = OrderedDict()
            TotalLoss = 0

            pred = net(data + eta.detach())

            loss = criterion(pred, label)
            TotalLoss = TotalLoss + loss
            wgrad = net.conv1.weight.grad
            # bgrad = net.conv1.bias.grad
            TotalLoss.backward()
            net.conv1.weight.grad = wgrad
            # net.conv1.bias.grad = bgrad
            # param = next(net.parameters())
            # grad_mean = torch.mean(param.grad)

            # optimizer.step()
            # optimizer.zero_grad()

            p = -1.0 * net.layer_one_out.grad
            yofo_inp, eta = LayerOneTrainner.step(data, p, eta)

            with torch.no_grad():
                if j == 0:
                    acc = torch_accuracy(pred, label, (1,))
                    cleanacc = acc[0].item()
                    cleanloss = loss.item()

                if j == K - 1:
                    yofo_pred = net(yofo_inp)
                    yofoacc = torch_accuracy(yofo_pred, label, (1,))[0].item()
            # pbar_dic['grad'] = '{}'.format(grad_mean)

        optimizer.step()
        LayerOneTrainner.param_optimizer.step()
        optimizer.zero_grad()
        LayerOneTrainner.param_optimizer.zero_grad()
        pbar_dic["Acc"] = f"{cleanacc:.2f}"
        pbar_dic["loss"] = f"{cleanloss:.2f}"
        pbar_dic["YofoAcc"] = f"{yofoacc:.2f}"
        pbar.set_postfix(pbar_dic)

    return cleanacc, yofoacc



import torch


@torch.no_grad()
def get_accuracy(model, data_loader, atk=None, n_limit=1e10, device=None):
    model = model.eval()

    if device is None:
        device = next(model.parameters()).device

    correct = 0
    total = 0

    for images, labels in data_loader:

        X = images.to(device)
        Y = labels.to(device)

        if atk:
            X = atk(X, Y)

        pre = model(X)

        _, pre = torch.max(pre.data, 1)
        total += pre.size(0)
        correct += (pre == Y).sum()

        if total > n_limit:
            break

    return 100 * float(correct) / total



import torch
import torch.nn as nn

# Fast is better than free (https://arxiv.org/abs/2001.03994)
class MNIST_Fast(nn.Module):
    def __init__(self, num_classes=10):
        super(MNIST_Fast, self).__init__()
        self.layers = nn.Sequential(
            nn.Conv2d(1, 16, 4, stride=2, padding=1),
            nn.ReLU(),
            nn.Conv2d(16, 32, 4, stride=2, padding=1),
            nn.ReLU(),
            nn.Flatten(),
            nn.Linear(32 * 7 * 7, 100),
            nn.ReLU(),
            nn.Linear(100, 10),
        )

    def forward(self, x):
        x = self.layers(x)
        return x



import torch
import torch.nn as nn
from torch.nn import functional as F


class LeNet(nn.Module):
    def __init__(self, num_classes=10):
        super().__init__()
        self.layers = nn.Sequential(
            nn.Conv2d(1, 32, 5),
            nn.ReLU(),
            nn.MaxPool2d(2, 2),
            nn.Conv2d(32, 64, 5),
            nn.ReLU(),
            nn.MaxPool2d(2, 2),
            nn.Flatten(),
            nn.Linear(64 * 4 * 4, 1024),
            nn.ReLU(),
            nn.Linear(1024, num_classes),
        )

    def forward(self, x):
        x = self.layers(x)
        return x


# https://github.com/KaiyangZhou/pytorch-center-loss/blob/master/models.py
class LeNetPP(nn.Module):
    def __init__(self, num_classes=10):
        super(LeNetPP, self).__init__()
        self.conv1_1 = nn.Conv2d(1, 32, 5, stride=1, padding=2)
        self.prelu1_1 = nn.PReLU()
        self.conv1_2 = nn.Conv2d(32, 32, 5, stride=1, padding=2)
        self.prelu1_2 = nn.PReLU()

        self.conv2_1 = nn.Conv2d(32, 64, 5, stride=1, padding=2)
        self.prelu2_1 = nn.PReLU()
        self.conv2_2 = nn.Conv2d(64, 64, 5, stride=1, padding=2)
        self.prelu2_2 = nn.PReLU()

        self.conv3_1 = nn.Conv2d(64, 128, 5, stride=1, padding=2)
        self.prelu3_1 = nn.PReLU()
        self.conv3_2 = nn.Conv2d(128, 128, 5, stride=1, padding=2)
        self.prelu3_2 = nn.PReLU()

        self.fc1 = nn.Linear(128 * 3 * 3, 2)
        self.prelu_fc1 = nn.PReLU()
        self.fc2 = nn.Linear(2, num_classes)

    def forward(self, x):
        x = self.prelu1_1(self.conv1_1(x))
        x = self.prelu1_2(self.conv1_2(x))
        x = F.max_pool2d(x, 2)

        x = self.prelu2_1(self.conv2_1(x))
        x = self.prelu2_2(self.conv2_2(x))
        x = F.max_pool2d(x, 2)

        x = self.prelu3_1(self.conv3_1(x))
        x = self.prelu3_2(self.conv3_2(x))
        x = F.max_pool2d(x, 2)

        x = x.view(-1, 128 * 3 * 3)
        x = self.prelu_fc1(self.fc1(x))
        y = self.fc2(x)

        return y



import setuptools

# read the contents of your README file
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


setuptools.setup(
    name="mair",
    version="1.0.3",
    description="MAIR is a PyTorch-based adversarial training framework.",
    author="Harry Kim",
    author_email="24k.harry@gmail.com",
    packages=setuptools.find_packages(),
    keyword=[
        "deeplearning",
        "adversarial",
        "attack",
        "pytorch",
        "torch",
        "defense",
        "at",
        "trades",
        "mart",
    ],
    install_requires=[
        "torch>=1.7.1",
        "torchvision>=0.8.2",
        "scipy>=0.14.0",
        "tqdm>=4.56.1",
        "requests~=2.25.1",
        "numpy>=1.19.4",
    ],
    python_requires=">=3",
    zip_safe=False,
    license="MIT",
    url="https://github.com/HarryK24/MAIR",
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 5 - Production/Stable",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Visualization",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: MIT License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3.6",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
)



# (c) 2024 LAiSR-SK

# This code is licensed under the MIT license (see LICENSE.md).
import os
from collections import namedtuple

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
import torchvision
from torchvision import transforms

import airtk.defense.gairat.gairat_attack_generator as attack
from airtk.model import ResNet18, WideResNet


class GairatTraining:
    def __init__(
        self,
        dataset: str = "cifar10",
        net: str = "resnet18",
        epochs: int = 120,
        weight_decay: float = 2e-4,
        momentum: float = 0.9,
        epsilon: float = 0.031,
        num_steps: int = 10,
        step_size=0.007,
        seed: int = 1,
        log_dir: str = "log/gairat_out.tx",
        lr_schedule: str = "piecewise",
        lr_max: float = 0.1,
        lr_one_drop: float = 0.1,
        lr_drop_epoch: int = 100,
        Lambda: int = -1,
        begin_epoch: int = 60,
        model_dir: str = "./data/model",
    ) -> None:
        ArgPrototype = namedtuple(
            "ArgPrototype",
            [
                "epochs",
                "weight_decay",
                "momentum",
                "epsilon",
                "num_steps",
                "step_size",
                "seed",
                "net",
                "dataset",
                "random",
                "depth",
                "width_factor",
                "drop_rate",
                "resume",
                "out_dir",
                "lr_schedule",
                "lr_max",
                "lr_one_drop",
                "lr_drop_epoch",
                "Lambda",
                "Lambda_max",
                "Lambda_schedule",
                "Weight_assignment_function",
                "begin_epoch",
                "model_dir",
            ],
        )

        self._args = ArgPrototype(
            epochs,
            weight_decay,
            momentum,
            epsilon,
            num_steps,
            step_size,
            seed,
            net,
            dataset,
            True,
            34,
            10,
            0.0,
            None,
            log_dir,
            lr_schedule,
            lr_max,
            lr_one_drop,
            lr_drop_epoch,
            Lambda,
            float("inf"),
            "fixed",
            "tanh",
            begin_epoch,
            model_dir,
        )

    def __call__(
        self,
    ) -> None:
        torch.manual_seed(self._args.seed)

        torch.backends.cudnn.benchmark = True

        torch.backends.cudnn.deterministic = True

        main_gairat(self._args)


# # Training settings

# seed = args.seed

# momentum = args.momentum

# weight_decay = args.weight_decay

# depth = args.depth

# width_factor = args.width_factor

# drop_rate = args.drop_rate

# resume = args.resume

# out_dir = args.out_dir


# Save checkpoint


def save_checkpoint(state, filename="checkpoint.pth.tar"):
    filepath = os.path.join(filename)

    torch.save(state, filepath)


# From GAIR.py file


def GAIR(num_steps, Kappa, Lambda, func):
    # Weight assign

    if func == "Tanh":
        reweight = (
            (
                Lambda
                + (int(num_steps / 2) - Kappa) * 5 / (int(num_steps / 2))
            ).tanh()
            + 1
        ) / 2

        normalized_reweight = reweight * len(reweight) / reweight.sum()

    elif func == "Sigmoid":
        reweight = (
            Lambda + (int(num_steps / 2) - Kappa) * 5 / (int(num_steps / 2))
        ).sigmoid()

        normalized_reweight = reweight * len(reweight) / reweight.sum()

    elif func == "Discrete":
        reweight = ((num_steps + 1) - Kappa) / (num_steps + 1)

        normalized_reweight = reweight * len(reweight) / reweight.sum()

    return normalized_reweight


# Get adversarially robust network


def train(args, epoch, model, train_loader, optimizer, Lambda, lr_schedule):
    lr = 0

    num_data = 0

    train_robust_loss = 0

    for batch_idx, (data, target) in enumerate(train_loader):
        loss = 0

        data, target = data.cuda(), target.cuda()

        # Get adversarial data and geometry value

        x_adv, Kappa = attack.GA_PGD(
            model,
            data,
            target,
            args.epsilon,
            args.step_size,
            args.num_steps,
            loss_fn="cent",
            category="Madry",
            rand_init=True,
        )

        model.train()

        lr = lr_schedule(epoch + 1)

        optimizer.param_groups[0].update(lr=lr)

        optimizer.zero_grad()

        logit = model(x_adv)

        if (epoch + 1) >= args.begin_epoch:
            Kappa = Kappa.cuda()

            loss = nn.CrossEntropyLoss(reduce=False)(logit, target)

            # Calculate weight assignment according to geometry value

            normalized_reweight = GAIR(
                args.num_steps, Kappa, Lambda, args.weight_assignment_function
            )

            loss = loss.mul(normalized_reweight).mean()

        else:
            loss = nn.CrossEntropyLoss(reduce="mean")(logit, target)

        train_robust_loss += loss.item() * len(x_adv)

        loss.backward()

        optimizer.step()

        num_data += len(data)

    train_robust_loss = train_robust_loss / num_data

    return train_robust_loss, lr


# Adjust lambda for weight assignment using epoch


def adjust_Lambda(epoch, args):
    Lam = float(args.Lambda)

    if args.epochs >= 110:
        # Train Wide-ResNet

        Lambda = args.Lambda_max

        if args.Lambda_schedule == "linear":
            if epoch >= 60:
                Lambda = args.Lambda_max - (epoch / args.epochs) * (
                    args.Lambda_max - Lam
                )

        elif args.Lambda_schedule == "piecewise":
            if epoch >= 60:
                Lambda = Lam

            elif epoch >= 90:
                Lambda = Lam - 1.0

            elif epoch >= 110:
                Lambda = Lam - 1.5

        elif args.Lambda_schedule == "fixed":
            if epoch >= 60:
                Lambda = Lam

    else:
        # Train ResNet

        Lambda = args.Lambda_max

        if args.Lambda_schedule == "linear":
            if epoch >= 30:
                Lambda = args.Lambda_max - (epoch / args.epochs) * (
                    args.Lambda_max - Lam
                )

        elif args.Lambda_schedule == "piecewise":
            if epoch >= 30:
                Lambda = Lam

            elif epoch >= 60:
                Lambda = Lam - 2.0

        elif args.Lambda_schedule == "fixed":
            if epoch >= 30:
                Lambda = Lam

    return Lambda


def main_gairat(args):
    # Training settings
    seed = args.seed
    momentum = args.momentum

    weight_decay = args.weight_decay
    depth = args.depth

    width_factor = args.width_factor
    drop_rate = args.drop_rate
    resume = args.resume
    out_dir = args.out_dir

    torch.manual_seed(seed)

    np.random.seed(seed)

    torch.cuda.manual_seed_all(seed)

    torch.backends.cudnn.benchmark = True

    torch.backends.cudnn.deterministic = True

    # Models and optimizer

    if args.net == "res18":
        model = ResNet18(
            num_classes=100 if args.dataset == "cifar100" else 10
        ).cuda()
    elif args.net == "wideres":
        nc = 100 if args.dataset == "cifar100" else 10

        model = WideResNet(
            depth=depth,
            num_classes=nc,
            widen_factor=width_factor,
            dropRate=drop_rate,
        ).cuda()
    else:
        raise NotImplementedError()

    model = torch.nn.DataParallel(model)

    optimizer = optim.SGD(
        model.parameters(),
        lr=args.lr_max,
        momentum=momentum,
        weight_decay=weight_decay,
    )

    # Learning schedules

    if args.lr_schedule == "superconverge":
        lr_schedule = lambda t: np.interp(
            [t], [0, args.epochs * 2 // 5, args.epochs], [0, args.lr_max, 0]
        )[0]

    elif args.lr_schedule == "piecewise":

        def lr_schedule(t):
            if args.epochs >= 110:
                # Train Wide-ResNet

                if t / args.epochs < 0.5:
                    return args.lr_max

                elif t / args.epochs < 0.75:
                    return args.lr_max / 10.0

                elif t / args.epochs < (11 / 12):
                    return args.lr_max / 100.0

                else:
                    return args.lr_max / 200.0

            else:
                # Train ResNet

                if t / args.epochs < 0.3:
                    return args.lr_max

                elif t / args.epochs < 0.6:
                    return args.lr_max / 10.0

                else:
                    return args.lr_max / 100.0

    elif args.lr_schedule == "linear":
        lr_schedule = lambda t: np.interp(
            [t],
            [0, args.epochs // 3, args.epochs * 2 // 3, args.epochs],
            [args.lr_max, args.lr_max, args.lr_max / 10, args.lr_max / 100],
        )[0]

    elif args.lr_schedule == "onedrop":

        def lr_schedule(t):
            if t < args.lr_drop_epoch:
                return args.lr_max

            else:
                return args.lr_one_drop

    elif args.lr_schedule == "multipledecay":

        def lr_schedule(t):
            return args.lr_max - (t // (args.epochs // 10)) * (
                args.lr_max / 10
            )

    elif args.lr_schedule == "cosine":

        def lr_schedule(t):
            return args.lr_max * 0.5 * (1 + np.cos(t / args.epochs * np.pi))

    # Store path

    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    # Setup data loader

    transform_train = transforms.Compose(
        [
            transforms.RandomCrop(32, padding=4),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
        ]
    )

    transform_test = transforms.Compose(
        [
            transforms.ToTensor(),
        ]
    )

    if args.dataset == "cifar10":
        trainset = torchvision.datasets.CIFAR10(
            root="./data/cifar-10",
            train=True,
            download=True,
            transform=transform_train,
        )

        train_loader = torch.utils.data.DataLoader(
            trainset, batch_size=128, shuffle=True, num_workers=2
        )

        testset = torchvision.datasets.CIFAR10(
            root="./data/cifar-10",
            train=False,
            download=True,
            transform=transform_test,
        )

        test_loader = torch.utils.data.DataLoader(
            testset, batch_size=128, shuffle=False, num_workers=2
        )

    if args.dataset == "cifar100":
        trainset = torchvision.datasets.CIFAR100(
            root="./data/cifar-100",
            train=True,
            download=True,
            transform=transform_train,
        )

        train_loader = torch.utils.data.DataLoader(
            trainset, batch_size=128, shuffle=True, num_workers=2
        )

        testset = torchvision.datasets.CIFAR100(
            root="./data/cifar-100",
            train=False,
            download=True,
            transform=transform_test,
        )

        test_loader = torch.utils.data.DataLoader(
            testset, batch_size=128, shuffle=False, num_workers=2
        )

    if args.dataset == "svhn":
        trainset = torchvision.datasets.SVHN(
            root="./data/SVHN",
            split="train",
            download=True,
            transform=transform_train,
        )

        train_loader = torch.utils.data.DataLoader(
            trainset, batch_size=128, shuffle=True, num_workers=2
        )

        testset = torchvision.datasets.SVHN(
            root="./data/SVHN",
            split="test",
            download=True,
            transform=transform_test,
        )

        test_loader = torch.utils.data.DataLoader(
            testset, batch_size=128, shuffle=False, num_workers=2
        )

    if args.dataset == "mnist":
        trainset = torchvision.datasets.MNIST(
            root="./data/MNIST",
            train=True,
            download=True,
            transform=transforms.ToTensor(),
        )

        train_loader = torch.utils.data.DataLoader(
            trainset,
            batch_size=128,
            shuffle=True,
            num_workers=1,
            pin_memory=True,
        )

        testset = torchvision.datasets.MNIST(
            root="./data/MNIST",
            train=False,
            download=True,
            transform=transforms.ToTensor(),
        )

        test_loader = torch.utils.data.DataLoader(
            testset,
            batch_size=128,
            shuffle=False,
            num_workers=1,
            pin_memory=True,
        )

    # Resume

    best_acc = 0

    start_epoch = 0

    if resume:
        # Resume directly point to checkpoint.pth.tar

        print("==> GAIRAT Resuming from checkpoint ..")

        print(resume)

        assert os.path.isfile(resume)

        out_dir = os.path.dirname(resume)

        checkpoint = torch.load(resume)

        start_epoch = checkpoint["epoch"]

        best_acc = checkpoint["test_pgd20_acc"]

        model.load_state_dict(checkpoint["state_dict"])

        optimizer.load_state_dict(checkpoint["optimizer"])

        # logger_test = Logger(os.path.join(out_dir, 'log_results.txt'), title=title, resume=True)

    else:
        print("==> GAIRAT")

        # logger_test = Logger(os.path.join(out_dir, 'log_results.txt'), title=title)

        # logger_test.set_names(['Epoch', 'Natural Test Acc', 'PGD20 Acc'])

    ## Training get started

    test_nat_acc = 0

    test_pgd20_acc = 0

    for epoch in range(start_epoch, args.epochs):
        # Get lambda

        Lambda = adjust_Lambda(epoch + 1, args)

        # Adversarial training

        train_robust_loss, lr = train(
            args, epoch, model, train_loader, optimizer, Lambda, lr_schedule
        )

        # Evalutions similar to DAT.

        _, test_nat_acc = attack.eval_clean(model, test_loader)

        _, test_pgd20_acc = attack.eval_robust(
            model,
            test_loader,
            perturb_steps=20,
            epsilon=0.031,
            step_size=0.031 / 4,
            loss_fn="cent",
            category="Madry",
            random=True,
        )

        print(
            "Epoch: [%d | %d] | Learning Rate: %f | Natural Test Acc %.2f | PGD20 Test Acc %.2f |\n"
            % (epoch, args.epochs, lr, test_nat_acc, test_pgd20_acc)
        )

        # logger_test.append([epoch + 1, test_nat_acc, test_pgd20_acc])

        # Save the best checkpoint

        if test_pgd20_acc > best_acc:
            best_acc = test_pgd20_acc

            save_checkpoint(
                {
                    "epoch": epoch + 1,
                    "state_dict": model.state_dict(),
                    "test_nat_acc": test_nat_acc,
                    "test_pgd20_acc": test_pgd20_acc,
                    "optimizer": optimizer.state_dict(),
                },
                filename=args.model_dir + "/bestpoint.pth.tar",
            )

        # Save the last checkpoint

        save_checkpoint(
            {
                "epoch": epoch + 1,
                "state_dict": model.state_dict(),
                "test_nat_acc": test_nat_acc,
                "test_pgd20_acc": test_pgd20_acc,
                "optimizer": optimizer.state_dict(),
            },
            filename=args.model_dir + f"/epoch-{epoch}.pth.tar",
        )

    # logger_test.close()

    # Save the final model

    model_dir = args.model_dir

    torch.save(
        model.state_dict(),
        os.path.join(model_dir, f"model-gairat-{args.dataset}-{args.net}.pt"),
    )



import torch
import torch.nn as nn
import torch.nn.functional as F

from ..attack import Attack


class TPGD(Attack):
    r"""
    PGD based on KL-Divergence loss in the paper 'Theoretically Principled Trade-off between Robustness and Accuracy'
    [https://arxiv.org/abs/1901.08573]

    Distance Measure : Linf

    Arguments:
        model (nn.Module): model to attack.
        eps (float): strength of the attack or maximum perturbation. (Default: 8/255)
        alpha (float): step size. (Default: 2/255)
        steps (int): number of steps. (Default: 10)

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.TPGD(model, eps=8/255, alpha=2/255, steps=10)
        >>> adv_images = attack(images)

    """

    def __init__(self, model, eps=8 / 255, alpha=2 / 255, steps=10):
        super().__init__("TPGD", model)
        self.eps = eps
        self.alpha = alpha
        self.steps = steps
        self.supported_mode = ["default"]

    def forward(self, images, labels=None):
        r"""
        Overridden.
        """

        images = images.clone().detach().to(self.device)
        logit_ori = self.get_logits(images).detach()

        adv_images = images + 0.001 * torch.randn_like(images)
        adv_images = torch.clamp(adv_images, min=0, max=1).detach()

        loss = nn.KLDivLoss(reduction="sum")

        for _ in range(self.steps):
            adv_images.requires_grad = True
            logit_adv = self.get_logits(adv_images)

            # Calculate loss
            cost = loss(F.log_softmax(logit_adv, dim=1), F.softmax(logit_ori, dim=1))

            # Update adversarial images
            grad = torch.autograd.grad(
                cost, adv_images, retain_graph=False, create_graph=False
            )[0]

            adv_images = adv_images.detach() + self.alpha * grad.sign()
            delta = torch.clamp(adv_images - images, min=-self.eps, max=self.eps)
            adv_images = torch.clamp(images + delta, min=0, max=1).detach()

        return adv_images



import torch
import torch.nn as nn

from ..attack import Attack


class EADL1(Attack):
    r"""
    EAD attack in the paper 'EAD: Elastic-Net Attacks to Deep Neural Networks'
    [https://arxiv.org/abs/1709.04114]

    Distance Measure : L1

    Arguments:
        model (nn.Module): model to attack.
        kappa (float): how strong the adversarial example should be (also written as 'confidence'). (Default: 0)
        lr (float): larger values converge faster to less accurate results. (Default: 0.01)
        binary_search_steps (int): number of times to adjust the constant with binary search. (Default: 9)
        max_iterations (int): number of iterations to perform gradient descent. (Default: 100)
        abort_early (bool): if we stop improving, abort gradient descent early. (Default: True)
        initial_const (float): the initial constant c to pick as a first guess. (Default: 0.001)
        beta (float): hyperparameter trading off L2 minimization for L1 minimization. (Default: 0.001)

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.EADL1(model, kappa=0, lr=0.01, max_iterations=100)
        >>> adv_images = attack(images, labels)

    """

    def __init__(
        self,
        model,
        kappa=0,
        lr=0.01,
        binary_search_steps=9,
        max_iterations=100,
        abort_early=True,
        initial_const=0.001,
        beta=0.001,
    ):
        super().__init__("EADL1", model)
        self.kappa = kappa
        self.lr = lr
        self.binary_search_steps = binary_search_steps
        self.max_iterations = max_iterations
        self.abort_early = abort_early
        self.initial_const = initial_const
        self.beta = beta
        # The last iteration (if we run many steps) repeat the search once.
        self.repeat = binary_search_steps >= 10
        self.supported_mode = ["default", "targeted"]

    def forward(self, images, labels):
        r"""
        Overridden.
        """

        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        if self.targeted:
            labels = self.get_target_label(images, labels)

        outputs = self.get_logits(images)

        batch_size = images.shape[0]
        lower_bound = torch.zeros(batch_size, device=self.device)
        const = torch.ones(batch_size, device=self.device) * self.initial_const
        upper_bound = torch.ones(batch_size, device=self.device) * 1e10

        final_adv_images = images.clone()
        y_one_hot = torch.eye(outputs.shape[1]).to(self.device)[labels]

        o_bestl1 = [1e10] * batch_size
        o_bestscore = [-1] * batch_size
        o_bestl1 = torch.Tensor(o_bestl1).to(self.device)
        o_bestscore = torch.Tensor(o_bestscore).to(self.device)

        # Initialization: x^{(0)} = y^{(0)} = x_0 in paper Algorithm 1 part
        x_k = images.clone().detach()
        y_k = nn.Parameter(images)

        # Start binary search
        for outer_step in range(self.binary_search_steps):
            self.global_step = 0

            bestl1 = [1e10] * batch_size
            bestscore = [-1] * batch_size

            bestl1 = torch.Tensor(bestl1).to(self.device)
            bestscore = torch.Tensor(bestscore).to(self.device)
            prevloss = 1e6

            if self.repeat and outer_step == (self.binary_search_steps - 1):
                const = upper_bound

            lr = self.lr
            for iteration in range(self.max_iterations):
                # reset gradient
                if y_k.grad is not None:
                    y_k.grad.detach_()
                    y_k.grad.zero_()

                # Loss over images_parameters with only L2 same as CW
                # we don't update L1 loss with SGD because we use ISTA
                output = self.get_logits(y_k)
                L2_loss = self.L2_loss(y_k, images)

                cost = self.EAD_loss(output, y_one_hot, None, L2_loss, const)
                # cost.backward(retain_graph=True)
                cost.backward()

                # Gradient step
                # y_k.data.add_(-lr, y_k.grad.data)
                self.global_step += 1
                with torch.no_grad():
                    y_k -= y_k.grad * lr

                # Ploynomial decay of learning rate
                lr = (
                    self.lr
                    * (1 - self.global_step / self.max_iterations) ** 0.5
                )  # nopep8
                x_k, y_k = self.FISTA(images, x_k, y_k)
                # Loss ElasticNet or L1 over x_k
                with torch.no_grad():
                    output = self.get_logits(x_k)
                    L2_loss = self.L2_loss(x_k, images)
                    L1_loss = self.L1_loss(x_k, images)
                    loss = self.EAD_loss(
                        output, y_one_hot, L1_loss, L2_loss, const
                    )  # nopep8

                    # print('loss: {}, prevloss: {}'.format(loss, prevloss))
                    if (
                        self.abort_early
                        and iteration % (self.max_iterations // 10) == 0
                    ):
                        if loss > prevloss * 0.999999:
                            break
                        prevloss = loss

                    # L1 attack key step!
                    cost = L1_loss
                    self.adjust_best_result(
                        x_k,
                        labels,
                        output,
                        cost,
                        bestl1,
                        bestscore,
                        o_bestl1,
                        o_bestscore,
                        final_adv_images,
                    )

            self.adjust_constant(
                labels, bestscore, const, upper_bound, lower_bound
            )

        return final_adv_images

    def L1_loss(self, x1, x2):
        Flatten = nn.Flatten()
        L1_loss = torch.abs(Flatten(x1) - Flatten(x2)).sum(dim=1)
        # L1_loss = L1.sum()
        return L1_loss

    def L2_loss(self, x1, x2):
        MSELoss = nn.MSELoss(reduction="none")
        Flatten = nn.Flatten()
        L2_loss = MSELoss(Flatten(x1), Flatten(x2)).sum(dim=1)
        # L2_loss = L2.sum()
        return L2_loss

    def EAD_loss(self, output, one_hot_labels, L1_loss, L2_loss, const):
        # Not same as CW's f function
        other = torch.max(
            (1 - one_hot_labels) * output - (one_hot_labels * 1e4), dim=1
        )[0]
        real = torch.max(one_hot_labels * output, dim=1)[0]

        if self.targeted:
            F_loss = torch.clamp((other - real), min=-self.kappa)
        else:
            F_loss = torch.clamp((real - other), min=-self.kappa)

        if isinstance(L1_loss, type(None)):
            loss = torch.sum(const * F_loss) + torch.sum(L2_loss)
        else:
            loss = (
                torch.sum(const * F_loss)
                + torch.sum(L2_loss)
                + torch.sum(self.beta * L1_loss)
            )

        return loss

    def FISTA(self, images, x_k, y_k):
        zt = self.global_step / (self.global_step + 3)

        upper = torch.clamp(y_k - self.beta, max=1)
        lower = torch.clamp(y_k + self.beta, min=0)

        diff = y_k - images
        cond1 = (diff > self.beta).float()
        cond2 = (torch.abs(diff) <= self.beta).float()
        cond3 = (diff < -self.beta).float()

        new_x_k = (cond1 * upper) + (cond2 * images) + (cond3 * lower)
        y_k.data = new_x_k + (zt * (new_x_k - x_k))
        return new_x_k, y_k

    def compare(self, output, labels):
        if len(output.shape) >= 2:
            # output is tensor
            output = output.clone().detach()
            if self.targeted:
                output[:, labels] -= self.kappa
            else:
                output[:, labels] += self.kappa
            output = torch.argmax(output, 1)
        else:
            # output is int or float
            pass

        if self.targeted:
            return output == labels
        else:
            return output != labels

    def adjust_best_result(
        self,
        adv_img,
        labels,
        output,
        cost,
        bestl1,
        bestscore,
        o_bestl1,
        o_bestscore,
        final_adv_images,
    ):
        output_label = torch.argmax(output, 1).float()
        mask = (cost < bestl1) & self.compare(output, labels)
        bestl1[mask] = cost[mask]
        bestscore[mask] = output_label[mask]

        mask = (cost < o_bestl1) & self.compare(output, labels)
        o_bestl1[mask] = cost[mask]
        o_bestscore[mask] = output_label[mask]
        final_adv_images[mask] = adv_img[mask]

    def adjust_constant(
        self, labels, bestscore, const, upper_bound, lower_bound
    ):
        mask = (self.compare(bestscore, labels)) & (bestscore != -1)
        upper_bound[mask] = torch.min(upper_bound[mask], const[mask])
        lower_bound[~mask] = torch.max(
            lower_bound[~mask], const[~mask]
        )  # nopep8

        mask = upper_bound < 1e9
        const[mask] = (lower_bound[mask] + upper_bound[mask]) / 2
        const[~mask] = const[~mask] * 10



import random
import numpy as np

import torch

# Modified from https://hoya012.github.io/
def manual_seed(random_seed=0):
    random_seed = 617
    torch.manual_seed(random_seed)
    torch.cuda.manual_seed(random_seed)
    torch.cuda.manual_seed_all(random_seed)  # if use multi-GPU
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    np.random.seed(random_seed)
    random.seed(random_seed)
    print("Fixed Random Seed:", random_seed)



import argparse
import logging
import math
import os
import time

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from airtk.defense.atawp.at_awp_utils import (
    Crop,
    Cutout,
    FlipLR,
    Transform,
    cifar10,
    cifar10_mean,
    cifar10_std,
    pad,
    transpose,
)
from airtk.defense.atawp.at_awp_utils_awp import (
    AdvWeightPerturb,
)
from airtk.model import ResNet18, WideResNet
from torch.autograd import Variable

mu = torch.tensor(cifar10_mean).view(3, 1, 1).cuda()

std = torch.tensor(cifar10_std).view(3, 1, 1).cuda()

device = torch.device("cuda")


def normalize(X):
    return (X - mu) / std


def clamp(X, lower_limit, upper_limit):
    return torch.max(torch.min(X, upper_limit), lower_limit)


class Batches:
    def __init__(
        self,
        dataset,
        batch_size,
        shuffle,
        set_random_choices=False,
        num_workers=0,
        drop_last=False,
    ):
        self.dataset = dataset

        self.batch_size = batch_size

        self.set_random_choices = set_random_choices

        self.dataloader = torch.utils.data.DataLoader(
            dataset,
            batch_size=batch_size,
            num_workers=num_workers,
            pin_memory=True,
            shuffle=shuffle,
            drop_last=drop_last,
        )

    def __iter__(self):
        if self.set_random_choices:
            self.dataset.set_random_choices()
        return (
            {"input": x.to(device).float(), "target": y.to(device).long()}
            for (x, y) in self.dataloader
        )

    def __len__(self):
        return len(self.dataloader)


def mixup_data(x, y, alpha=1.0):
    """Returns mixed inputs, pairs of targets, and lambda"""

    lam = np.random.beta(alpha, alpha) if alpha > 0 else 1

    batch_size = x.size()[0]

    index = torch.randperm(batch_size).cuda()

    mixed_x = lam * x + (1 - lam) * x[index, :]

    y_a, y_b = y, y[index]

    return mixed_x, y_a, y_b, lam


def mixup_criterion(criterion, pred, y_a, y_b, lam):
    return lam * criterion(pred, y_a) + (1 - lam) * criterion(pred, y_b)


def attack_pgd(
    model,
    X,
    y,
    epsilon,
    alpha,
    attack_iters,
    restarts,
    norm,
    early_stop=False,
    mixup=False,
    y_a=None,
    y_b=None,
    lam=None,
):
    max_loss = torch.zeros(y.shape[0]).cuda()

    max_delta = torch.zeros_like(X).cuda()

    lower_limit = 0

    upper_limit = 1

    for _ in range(restarts):
        delta = torch.zeros_like(X).cuda()

        if norm == "l_inf":
            delta.uniform_(-epsilon, epsilon)

        elif norm == "l_2":
            delta.normal_()

            d_flat = delta.view(delta.size(0), -1)

            n = d_flat.norm(p=2, dim=1).view(delta.size(0), 1, 1, 1)

            r = torch.zeros_like(n).uniform_(0, 1)

            delta *= r / n * epsilon

        else:
            raise ValueError

        delta = clamp(delta, lower_limit - X, upper_limit - X)

        delta.requires_grad = True

        for _ in range(attack_iters):
            output = model(normalize(X + delta))

            if early_stop:
                index = torch.where(output.max(1)[1] == y)[0]

            else:
                index = slice(None, None, None)

            if not isinstance(index, slice) and len(index) == 0:
                break

            if mixup:
                criterion = nn.CrossEntropyLoss()

                loss = mixup_criterion(
                    criterion, model(normalize(X + delta)), y_a, y_b, lam
                )

            else:
                loss = F.cross_entropy(output, y)

            loss.backward()

            grad = delta.grad.detach()

            d = delta[index, :, :, :]

            g = grad[index, :, :, :]

            x = X[index, :, :, :]

            if norm == "l_inf":
                d = torch.clamp(
                    d + alpha * torch.sign(g), min=-epsilon, max=epsilon
                )

            elif norm == "l_2":
                g_norm = torch.norm(g.view(g.shape[0], -1), dim=1).view(
                    -1, 1, 1, 1
                )

                scaled_g = g / (g_norm + 1e-10)
                d = (
                    (d + scaled_g * alpha)
                    .view(d.size(0), -1)
                    .renorm(p=2, dim=0, maxnorm=epsilon)
                    .view_as(d)
                )

            d = clamp(d, lower_limit - x, upper_limit - x)

            delta.data[index, :, :, :] = d

            delta.grad.zero_()

        if mixup:
            criterion = nn.CrossEntropyLoss(reduction="none")

            all_loss = mixup_criterion(
                criterion, model(normalize(X + delta)), y_a, y_b, lam
            )

        else:
            all_loss = F.cross_entropy(
                model(normalize(X + delta)), y, reduction="none"
            )

        max_delta[all_loss >= max_loss] = delta.detach()[all_loss >= max_loss]

        max_loss = torch.max(max_loss, all_loss)

    return max_delta


def get_args():
    parser = argparse.ArgumentParser()

    parser.add_argument("--model", default="ResNet18")

    parser.add_argument("--l2", default=0, type=float)

    parser.add_argument("--l1", default=0, type=float)

    parser.add_argument("--batch-size", default=128, type=int)

    parser.add_argument("--batch-size-test", default=128, type=int)

    parser.add_argument("--data-dir", default="../cifar-data", type=str)

    parser.add_argument("--epochs", default=200, type=int)
    parser.add_argument(
        "--lr-schedule",
        default="piecewise",
        choices=[
            "superconverge",
            "piecewise",
            "linear",
            "piecewisesmoothed",
            "piecewisezoom",
            "onedrop",
            "multipledecay",
            "cosine",
            "cyclic",
        ],
    )

    parser.add_argument("--lr-max", default=0.1, type=float)

    parser.add_argument("--lr-one-drop", default=0.01, type=float)

    parser.add_argument("--lr-drop-epoch", default=100, type=int)
    parser.add_argument(
        "--attack",
        default="pgd",
        type=str,
        choices=["pgd", "fgsm", "free", "none"],
    )

    parser.add_argument("--epsilon", default=8, type=int)

    parser.add_argument("--attack-iters", default=10, type=int)

    parser.add_argument("--attack-iters-test", default=20, type=int)

    parser.add_argument("--restarts", default=1, type=int)

    parser.add_argument("--pgd-alpha", default=2, type=float)

    parser.add_argument("--fgsm-alpha", default=1.25, type=float)
    parser.add_argument(
        "--norm", default="l_inf", type=str, choices=["l_inf", "l_2"]
    )
    parser.add_argument(
        "--fgsm-init", default="random", choices=["zero", "random", "previous"]
    )

    parser.add_argument("--fname", default="cifar_model", type=str)

    parser.add_argument("--seed", default=0, type=int)

    parser.add_argument("--half", action="store_true")

    parser.add_argument("--width-factor", default=10, type=int)

    parser.add_argument("--resume", default=0, type=int)

    parser.add_argument("--cutout", action="store_true")

    parser.add_argument("--cutout-len", type=int)

    parser.add_argument("--mixup", action="store_true")

    parser.add_argument("--mixup-alpha", type=float)

    parser.add_argument("--eval", action="store_true")

    parser.add_argument("--val", action="store_true")

    parser.add_argument("--chkpt-iters", default=10, type=int)

    parser.add_argument("--awp-gamma", default=0.01, type=float)

    parser.add_argument("--awp-warmup", default=0, type=int)
    return parser.parse_args()


def main():
    args = get_args()

    lower_limit = 0

    upper_limit = 1

    if args.awp_gamma <= 0.0:
        args.awp_warmup = np.infty

    if not os.path.exists(args.fname):
        os.makedirs(args.fname)

    logger = logging.getLogger(__name__)

    logging.basicConfig(
        format="[%(asctime)s] - %(message)s",
        datefmt="%Y/%m/%d %H:%M:%S",
        level=logging.DEBUG,
        handlers=[
            logging.FileHandler(
                os.path.join(
                    args.fname, "eval.log" if args.eval else "output.log"
                )
            ),
            logging.StreamHandler(),
        ],
    )

    logger.info(args)

    np.random.seed(args.seed)

    torch.manual_seed(args.seed)

    torch.cuda.manual_seed(args.seed)

    transforms = [Crop(32, 32), FlipLR()]

    if args.cutout:
        transforms.append(Cutout(args.cutout_len, args.cutout_len))

    if args.val:
        try:
            dataset = torch.load("cifar10_validation_split.pth")

        except Exception:
            print(
                "Couldn't find a dataset with a validation split, did you run "
                "generate_validation.py?"
            )
            return

        val_set = list(
            zip(
                transpose(dataset["val"]["data"] / 255.0),
                dataset["val"]["labels"],
            )
        )

        val_batches = Batches(
            val_set, args.batch_size, shuffle=False, num_workers=2
        )

    else:
        dataset = cifar10(args.data_dir)
    train_set = list(
        zip(
            transpose(pad(dataset["train"]["data"], 4) / 255.0),
            dataset["train"]["labels"],
        )
    )

    train_set_x = Transform(train_set, transforms)

    train_batches = Batches(
        train_set_x,
        args.batch_size,
        shuffle=True,
        set_random_choices=True,
        num_workers=2,
    )

    test_set = list(
        zip(
            transpose(dataset["test"]["data"] / 255.0),
            dataset["test"]["labels"],
        )
    )

    test_batches = Batches(
        test_set, args.batch_size_test, shuffle=False, num_workers=2
    )

    epsilon = args.epsilon / 255.0

    pgd_alpha = args.pgd_alpha / 255.0

    if args.model == "ResNet18":
        model = ResNet18(num_classes=10)

        proxy = ResNet18(num_classes=10)

    elif args.model == "WideResNet":
        model = WideResNet(depth=34, num_classes=10).to(device)

        proxy = WideResNet(depth=34, num_classes=10).to(device)

    else:
        raise ValueError("Unknown model")

    model = nn.DataParallel(model).cuda()

    proxy = nn.DataParallel(proxy).cuda()

    if args.l2:
        decay, no_decay = [], []

        for name, param in model.named_parameters():
            if "bn" not in name and "bias" not in name:
                decay.append(param)

            else:
                no_decay.append(param)

        params = [
            {"params": decay, "weight_decay": args.l2},
            {"params": no_decay, "weight_decay": 0},
        ]

    else:
        params = model.parameters()

    opt = torch.optim.SGD(
        params, lr=args.lr_max, momentum=0.9, weight_decay=5e-4
    )

    proxy_opt = torch.optim.SGD(proxy.parameters(), lr=0.01)

    awp_adversary = AdvWeightPerturb(
        model=model, proxy=proxy, proxy_optim=proxy_opt, gamma=args.awp_gamma
    )

    criterion = nn.CrossEntropyLoss()

    if (
        args.attack == "free"
        or args.attack == "fgsm"
        and args.fgsm_init == "previous"
    ):
        delta = torch.zeros(args.batch_size, 3, 32, 32).cuda()

        delta.requires_grad = True

    if args.attack == "free":
        epochs = int(math.ceil(args.epochs / args.attack_iters))

    else:
        epochs = args.epochs

    if args.lr_schedule == "superconverge":
        lr_schedule = lambda t: np.interp(
            [t], [0, args.epochs * 2 // 5, args.epochs], [0, args.lr_max, 0]
        )[0]

    elif args.lr_schedule == "piecewise":

        def lr_schedule(t):
            if t / args.epochs < 0.5:
                return args.lr_max

            elif t / args.epochs < 0.75:
                return args.lr_max / 10.0

            else:
                return args.lr_max / 100.0

    elif args.lr_schedule == "linear":
        lr_schedule = lambda t: np.interp(
            [t],
            [0, args.epochs // 3, args.epochs * 2 // 3, args.epochs],
            [args.lr_max, args.lr_max, args.lr_max / 10, args.lr_max / 100],
        )[0]

    elif args.lr_schedule == "onedrop":

        def lr_schedule(t):
            if t < args.lr_drop_epoch:
                return args.lr_max

            else:
                return args.lr_one_drop

    elif args.lr_schedule == "multipledecay":

        def lr_schedule(t):
            return args.lr_max - (t // (args.epochs // 10)) * (
                args.lr_max / 10
            )

    elif args.lr_schedule == "cosine":

        def lr_schedule(t):
            return args.lr_max * 0.5 * (1 + np.cos(t / args.epochs * np.pi))

    elif args.lr_schedule == "cyclic":
        lr_schedule = lambda t: np.interp(
            [t], [0, 0.4 * args.epochs, args.epochs], [0, args.lr_max, 0]
        )[0]

    best_test_robust_acc = 0

    best_val_robust_acc = 0

    if args.resume:
        start_epoch = args.resume
        model.load_state_dict(
            torch.load(os.path.join(args.fname, f"model_{start_epoch-1}.pth"))
        )
        opt.load_state_dict(
            torch.load(os.path.join(args.fname, f"opt_{start_epoch-1}.pth"))
        )

        logger.info(f"Resuming at epoch {start_epoch}")

        if os.path.exists(os.path.join(args.fname, "model_best.pth")):
            best_test_robust_acc = torch.load(
                os.path.join(args.fname, "model_best.pth")
            )["test_robust_acc"]

        if args.val:
            best_val_robust_acc = torch.load(
                os.path.join(args.fname, "model_val.pth")
            )["val_robust_acc"]

    else:
        start_epoch = 0

    if args.eval:
        if not args.resume:
            logger.info(
                "No model loaded to evaluate, specify with --resume FNAME"
            )
            return

        logger.info("[Evaluation mode]")

    logger.info(
        "Epoch \t Train Time \t Test Time \t LR \t \t Train Loss \t Train Acc \t Train Robust Loss \t Train Robust Acc \t Test Loss \t Test Acc \t Test Robust Loss \t Test Robust Acc"
    )

    for epoch in range(start_epoch, epochs):
        start_time = time.time()

        train_loss = 0

        train_acc = 0

        train_robust_loss = 0

        train_robust_acc = 0

        train_n = 0

        for i, batch in enumerate(train_batches):
            if args.eval:
                break

            X, y = batch["input"], batch["target"]

            if args.mixup:
                X, y_a, y_b, lam = mixup_data(X, y, args.mixup_alpha)

                X, y_a, y_b = map(Variable, (X, y_a, y_b))

            lr = lr_schedule(epoch + (i + 1) / len(train_batches))

            opt.param_groups[0].update(lr=lr)

            if args.attack == "pgd":
                # Random initialization

                if args.mixup:
                    delta = attack_pgd(
                        model,
                        X,
                        y,
                        epsilon,
                        pgd_alpha,
                        args.attack_iters,
                        args.restarts,
                        args.norm,
                        mixup=True,
                        y_a=y_a,
                        y_b=y_b,
                        lam=lam,
                    )

                else:
                    delta = attack_pgd(
                        model,
                        X,
                        y,
                        epsilon,
                        pgd_alpha,
                        args.attack_iters,
                        args.restarts,
                        args.norm,
                    )

                delta = delta.detach()

            elif args.attack == "fgsm":
                delta = attack_pgd(
                    model,
                    X,
                    y,
                    epsilon,
                    args.fgsm_alpha * epsilon,
                    1,
                    1,
                    args.norm,
                )

            # Standard training

            elif args.attack == "none":
                delta = torch.zeros_like(X)

            X_adv = normalize(
                torch.clamp(
                    X + delta[: X.size(0)], min=lower_limit, max=upper_limit
                )
            )

            model.train()

            # calculate adversarial weight perturbation and perturb it

            if epoch >= args.awp_warmup:
                # not compatible to mixup currently.

                assert not args.mixup

                awp = awp_adversary.calc_awp(inputs_adv=X_adv, targets=y)

                awp_adversary.perturb(awp)

            robust_output = model(X_adv)

            if args.mixup:
                robust_loss = mixup_criterion(
                    criterion, robust_output, y_a, y_b, lam
                )

            else:
                robust_loss = criterion(robust_output, y)

            if args.l1:
                for name, param in model.named_parameters():
                    if "bn" not in name and "bias" not in name:
                        robust_loss += args.l1 * param.abs().sum()

            opt.zero_grad()

            robust_loss.backward()
            opt.step()

            if epoch >= args.awp_warmup:
                awp_adversary.restore(awp)

            output = model(normalize(X))

            if args.mixup:
                loss = mixup_criterion(criterion, output, y_a, y_b, lam)

            else:
                loss = criterion(output, y)

            train_robust_loss += robust_loss.item() * y.size(0)

            train_robust_acc += (robust_output.max(1)[1] == y).sum().item()

            train_loss += loss.item() * y.size(0)

            train_acc += (output.max(1)[1] == y).sum().item()

            train_n += y.size(0)

        train_time = time.time()

        model.eval()

        test_loss = 0

        test_acc = 0

        test_robust_loss = 0

        test_robust_acc = 0

        test_n = 0

        for batch in test_batches:
            X, y = batch["input"], batch["target"]

            # Random initialization

            if args.attack == "none":
                delta = torch.zeros_like(X)

            else:
                delta = attack_pgd(
                    model,
                    X,
                    y,
                    epsilon,
                    pgd_alpha,
                    args.attack_iters_test,
                    args.restarts,
                    args.norm,
                    early_stop=args.eval,
                )

            delta = delta.detach()

            robust_output = model(
                normalize(
                    torch.clamp(
                        X + delta[: X.size(0)],
                        min=lower_limit,
                        max=upper_limit,
                    )
                )
            )

            robust_loss = criterion(robust_output, y)

            output = model(normalize(X))

            loss = criterion(output, y)

            test_robust_loss += robust_loss.item() * y.size(0)

            test_robust_acc += (robust_output.max(1)[1] == y).sum().item()

            test_loss += loss.item() * y.size(0)

            test_acc += (output.max(1)[1] == y).sum().item()

            test_n += y.size(0)

        test_time = time.time()

        if args.val:
            val_loss = 0

            val_acc = 0

            val_robust_loss = 0

            val_robust_acc = 0

            val_n = 0

            for i, batch in enumerate(val_batches):
                X, y = batch["input"], batch["target"]

                # Random initialization

                if args.attack == "none":
                    delta = torch.zeros_like(X)

                else:
                    delta = attack_pgd(
                        model,
                        X,
                        y,
                        epsilon,
                        pgd_alpha,
                        args.attack_iters_test,
                        args.restarts,
                        args.norm,
                        early_stop=args.eval,
                    )

                delta = delta.detach()

                robust_output = model(
                    normalize(
                        torch.clamp(
                            X + delta[: X.size(0)],
                            min=lower_limit,
                            max=upper_limit,
                        )
                    )
                )

                robust_loss = criterion(robust_output, y)

                output = model(normalize(X))

                loss = criterion(output, y)

                val_robust_loss += robust_loss.item() * y.size(0)

                val_robust_acc += (robust_output.max(1)[1] == y).sum().item()

                val_loss += loss.item() * y.size(0)

                val_acc += (output.max(1)[1] == y).sum().item()

                val_n += y.size(0)

        if not args.eval:
            logger.info(
                "%d \t %.1f \t \t %.1f \t \t %.4f \t %.4f \t %.4f \t %.4f \t \t %.4f \t \t %.4f \t %.4f \t %.4f \t \t %.4f",
                epoch,
                train_time - start_time,
                test_time - train_time,
                lr,
                train_loss / train_n,
                train_acc / train_n,
                train_robust_loss / train_n,
                train_robust_acc / train_n,
                test_loss / test_n,
                test_acc / test_n,
                test_robust_loss / test_n,
                test_robust_acc / test_n,
            )

            if args.val:
                logger.info(
                    "validation %.4f \t %.4f \t %.4f \t %.4f",
                    val_loss / val_n,
                    val_acc / val_n,
                    val_robust_loss / val_n,
                    val_robust_acc / val_n,
                )

                if val_robust_acc / val_n > best_val_robust_acc:
                    torch.save(
                        {
                            "state_dict": model.state_dict(),
                            "test_robust_acc": test_robust_acc / test_n,
                            "test_robust_loss": test_robust_loss / test_n,
                            "test_loss": test_loss / test_n,
                            "test_acc": test_acc / test_n,
                            "val_robust_acc": val_robust_acc / val_n,
                            "val_robust_loss": val_robust_loss / val_n,
                            "val_loss": val_loss / val_n,
                            "val_acc": val_acc / val_n,
                        },
                        os.path.join(args.fname, "model_val.pth"),
                    )

                    best_val_robust_acc = val_robust_acc / val_n

            # save checkpoint

            if (epoch + 1) % args.chkpt_iters == 0 or epoch + 1 == epochs:
                torch.save(
                    model.state_dict(),
                    os.path.join(args.fname, f"model_{epoch}.pth"),
                )

                torch.save(
                    opt.state_dict(),
                    os.path.join(args.fname, f"opt_{epoch}.pth"),
                )

            # save best

            if test_robust_acc / test_n > best_test_robust_acc:
                torch.save(
                    {
                        "state_dict": model.state_dict(),
                        "test_robust_acc": test_robust_acc / test_n,
                        "test_robust_loss": test_robust_loss / test_n,
                        "test_loss": test_loss / test_n,
                        "test_acc": test_acc / test_n,
                    },
                    os.path.join(args.fname, "model_best.pth"),
                )

                best_test_robust_acc = test_robust_acc / test_n

        else:
            logger.info(
                "%d \t %.1f \t \t %.1f \t \t %.4f \t %.4f \t %.4f \t %.4f \t \t %.4f \t \t %.4f \t %.4f \t %.4f \t \t %.4f",
                epoch,
                train_time - start_time,
                test_time - train_time,
                -1,
                -1,
                -1,
                -1,
                -1,
                test_loss / test_n,
                test_acc / test_n,
                test_robust_loss / test_n,
                test_robust_acc / test_n,
            )
            return model
    return model


def main_atawp10():
    model = main()

    model_dir = "../data/model"

    torch.save(
        model.state_dict(),
        os.path.join(
            model_dir, "model-at-awp-{}-{}.pt".format("cifar10", "res18")
        ),
    )

    target_model = ResNet18(num_classes=10)

    from collections import OrderedDict

    checkpoint = torch.load(
        os.path.join(
            model_dir, "model-at-awp-{}-{}.pt".format("cifar10", "res18")
        )
    )

    try:
        target_model.load_state_dict(checkpoint)

    except Exception:
        new_state_dict = OrderedDict()

        for k, v in checkpoint.items():
            name = k[7:]  # remove `module.`

            new_state_dict[name] = v

        target_model.load_state_dict(new_state_dict, False)

    target_model = torch.nn.DataParallel(target_model).cuda()

    torch.save(
        target_model.module.state_dict(),
        os.path.join(
            model_dir, "model-at-awp-{}-{}.pt".format("cifar10", "res18")
        ),
    )


if __name__ == "__main__":
    main()



from abc import ABCMeta, abstractmethod

import numpy as np
import torch


class AttackBase(metaclass=ABCMeta):
    @abstractmethod
    def attack(self, net, inp, label, target=None):
        """

        :param inp: batched images
        :param target: specify the indexes of target class, None represents untargeted attack
        :return: batched adversaril images
        """
        pass

    @abstractmethod
    def to(self, device):
        pass


def clip_eta(eta, norm, eps, DEVICE=torch.device("cuda:0")):
    """
    helper functions to project eta into epsilon norm ball
    :param eta: Perturbation tensor (should be of size(N, C, H, W))
    :param norm: which norm. should be in [1, 2, np.inf]
    :param eps: epsilon, bound of the perturbation
    :return: Projected perturbation
    """

    assert norm in [1, 2, np.inf], "norm should be in [1, 2, np.inf]"

    with torch.no_grad():
        avoid_zero_div = torch.tensor(1e-12).to(DEVICE)
        eps = torch.tensor(eps).to(DEVICE)
        one = torch.tensor(1.0).to(DEVICE)

        if norm == np.inf:
            eta = torch.clamp(eta, -eps, eps)
        else:
            normalize = torch.norm(
                eta.reshape(eta.size(0), -1), p=norm, dim=-1, keepdim=False
            )
            normalize = torch.max(normalize, avoid_zero_div)

            normalize.unsqueeze_(dim=-1)
            normalize.unsqueeze_(dim=-1)
            normalize.unsqueeze_(dim=-1)

            factor = torch.min(one, eps / normalize)
            eta = eta * factor
    return eta


def test_clip():
    a = torch.rand((10, 3, 28, 28)).cuda()

    epss = [0.1, 0.5, 1]

    norms = [1, 2, np.inf]
    for e, n in zip(epss, norms):
        print(e, n)
        c = clip_eta(a, n, e, True)

        print(c)


if __name__ == "__main__":
    test_clip()



import torch
import torch.nn as nn

from ..attack import Attack


class VNIFGSM(Attack):
    r"""
    VNI-FGSM in the paper 'Enhancing the Transferability of Adversarial Attacks through Variance Tuning
    [https://arxiv.org/abs/2103.15571], Published as a conference paper at CVPR 2021
    Modified from "https://github.com/JHL-HUST/VT"

    Distance Measure : Linf

    Arguments:
        model (nn.Module): model to attack.
        eps (float): maximum perturbation. (Default: 8/255)
        alpha (float): step size. (Default: 2/255)
        steps (int): number of iterations. (Default: 10)
        decay (float): momentum factor. (Default: 1.0)
        N (int): the number of sampled examples in the neighborhood. (Default: 5)
        beta (float): the upper bound of neighborhood. (Default: 3/2)

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.VNIFGSM(model, eps=8/255, alpha=2/255, steps=10, decay=1.0, N=5, beta=3/2)
        >>> adv_images = attack(images, labels)

    """

    def __init__(
        self, model, eps=8 / 255, alpha=2 / 255, steps=10, decay=1.0, N=5, beta=3 / 2
    ):
        super().__init__("VNIFGSM", model)
        self.eps = eps
        self.steps = steps
        self.decay = decay
        self.alpha = alpha
        self.N = N
        self.beta = beta
        self.supported_mode = ["default", "targeted"]

    def forward(self, images, labels):
        r"""
        Overridden.
        """

        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        if self.targeted:
            target_labels = self.get_target_label(images, labels)

        momentum = torch.zeros_like(images).detach().to(self.device)
        v = torch.zeros_like(images).detach().to(self.device)
        loss = nn.CrossEntropyLoss()
        adv_images = images.clone().detach()

        for _ in range(self.steps):
            adv_images.requires_grad = True
            nes_images = adv_images + self.decay * self.alpha * momentum
            outputs = self.get_logits(nes_images)

            # Calculate loss
            if self.targeted:
                cost = -loss(outputs, target_labels)
            else:
                cost = loss(outputs, labels)

            # Update adversarial images
            adv_grad = torch.autograd.grad(
                cost, adv_images, retain_graph=False, create_graph=False
            )[0]

            grad = (adv_grad + v) / torch.mean(
                torch.abs(adv_grad + v), dim=(1, 2, 3), keepdim=True
            )
            grad = grad + momentum * self.decay
            momentum = grad

            # Calculate Gradient Variance
            GV_grad = torch.zeros_like(images).detach().to(self.device)
            for _ in range(self.N):
                neighbor_images = adv_images.detach() + torch.randn_like(
                    images
                ).uniform_(-self.eps * self.beta, self.eps * self.beta)
                neighbor_images.requires_grad = True
                outputs = self.get_logits(neighbor_images)

                # Calculate loss
                if self.targeted:
                    cost = -loss(outputs, target_labels)
                else:
                    cost = loss(outputs, labels)
                GV_grad += torch.autograd.grad(
                    cost, neighbor_images, retain_graph=False, create_graph=False
                )[0]
            # obtaining the gradient variance
            v = GV_grad / self.N - adv_grad

            adv_images = adv_images.detach() + self.alpha * grad.sign()
            delta = torch.clamp(adv_images - images, min=-self.eps, max=self.eps)
            adv_images = torch.clamp(images + delta, min=0, max=1).detach()

        return adv_images



# (c) 2024 LAiSR-SK

# This code is licensed under the MIT license (see LICENSE.md).

import argparse
import copy
import os
import time

import torch
from airtk.helper_functions import (
    adjust_learning_rate,
    eval_clean,
    load_data,
    robust_eval,
)
from airtk.loss import adt_loss
from airtk.model import WideResNet
from torch import optim

parser = argparse.ArgumentParser(
    description="PyTorch Adversarial Training Framework: ADT"
)
parser.add_argument(
    "--batch-size",
    type=int,
    default=128,
    metavar="N",
    help="input batch size for training (default: 128)",
)
parser.add_argument(
    "--test-batch-size",
    type=int,
    default=128,
    metavar="N",
    help="input batch size for testing (default: 128)",
)
parser.add_argument(
    "--epochs",
    type=int,
    default=76,
    metavar="N",
    help="number of epochs to train",
)
parser.add_argument(
    "--weight-decay", "--wd", default=2e-4, type=float, metavar="W"
)
parser.add_argument(
    "--lr", type=float, default=0.1, metavar="LR", help="learning rate"
)
parser.add_argument(
    "--momentum", type=float, default=0.9, metavar="M", help="SGD momentum"
)
parser.add_argument(
    "--no-cuda",
    action="store_true",
    default=False,
    help="disables CUDA training",
)
parser.add_argument(
    "--seed", type=int, default=1, metavar="S", help="random seed (default: 1)"
)
parser.add_argument(
    "--log-interval",
    type=int,
    default=100,
    metavar="N",
    help="how many batches to wait before logging training status",
)
parser.add_argument(
    "--model-dir",
    default="./data/model",
    help="directory of model for saving checkpoint",
)
parser.add_argument(
    "--save-freq",
    "-s",
    default=1,
    type=int,
    metavar="N",
    help="save frequency",
)
parser.add_argument(
    "--lr-schedule",
    default="decay",
    help="schedule for adjusting learning rate",
)
# args = parser.parse_args()


# Establish the settings
model_dir = args.model_dir

if not os.path.exists(model_dir):
    os.makedirs(model_dir)

use_cuda = not args.no_cuda and torch.cuda.is_available()
torch.manual_seed(args.seed)

device = torch.device("cuda" if use_cuda else "cpu")

kwargs = {"num_workers": 1, "pin_memory": True} if use_cuda else {}


def train(args, model, device, train_loader, optimizer, ds_name, epoch):
    """

    Trains one epoch by calculating the loss for each batch and updating the model


    :param args: set of arguments including learning rate, log interval, etc

    :param model: model to train

    :param device: the current device

    :param train_loader: data loader containing the training dataset

    :param optimizer: optimizer to train

    :param epoch: current epoch of training
    """

    model.train()  # set the  model to training mode

    for batch_idx, sample in enumerate(train_loader):
        # Set the data and target (x and y) to the device

        if ds_name == "cifar100":
            data, target, coarse = sample

        else:
            data, target = sample

        data, target = data.to(device), target.to(device)

        optimizer.zero_grad()  # zero out the gradients

        # Calculate the ADT loss

        loss = adt_loss(model, data, target, optimizer)

        # Update the model based on the loss

        loss.backward()

        optimizer.step()

        # print progress

        if batch_idx % args.log_interval == 0:
            print(
                f"Train Epoch: {epoch} [{batch_idx * len(data)}/{len(train_loader.dataset)} ({100.0 * batch_idx / len(train_loader):.0f}%)]\tLoss: {loss.item():.6f}"
            )


def main_adt(ds_name, mod_name="wideres34"):
    """

    Main training method, which establishes the model/dataset before conducting training and

    clean testing for each epoch. Then reports final training time and conducts robustness tests
    on the final model.


    :param ds_name: the dataset to use for training

    :param mod_name: the model to use for training
    """

    # Set up file for printing the output

    filename = f"log/adt-{ds_name}-{mod_name}-output.txt"
    f = open(filename, "a")

    # Initialize the desired model

    start_tot = time.time()  # start timing the full method

    if mod_name == "res18":
        model = ResNet18(num_classes=100 if ds_name == "cifar100" else 10).to(
            device
        )

    elif mod_name == "res34":
        model = ResNet34(num_classes=100 if ds_name == "cifar100" else 10).to(
            device
        )

    elif mod_name == "res50":
        model = ResNet50(num_classes=100 if ds_name == "cifar100" else 10).to(
            device
        )

    elif mod_name == "res101":
        model = ResNet101(num_classes=100 if ds_name == "cifar100" else 10).to(
            device
        )

    elif mod_name == "res152":
        model = ResNet152(num_classes=100 if ds_name == "cifar100" else 10).to(
            device
        )

    elif mod_name == "wideres34":
        model = WideResNet(
            depth=34, num_classes=100 if ds_name == "cifar100" else 10
        ).to(device)

    else:
        raise NotImplementedError

    # Set up the optimizer to the arguments

    optimizer = optim.SGD(
        model.parameters(),
        lr=args.lr,
        momentum=args.momentum,
        weight_decay=args.weight_decay,
    )

    # Set up the dataloaders

    train_loader, test_loader = load_data(ds_name, args, kwargs, coarse=True)

    # Save the model and optimizer

    torch.save(
        model.state_dict(),
        os.path.join(model_dir, f"model-adt-{ds_name}-{mod_name}-start.pt"),
    )

    torch.save(
        optimizer.state_dict(),
        os.path.join(model_dir, f"opt-adt-{ds_name}-{mod_name}-start.tar"),
    )

    # Begin training for the designated number of epochs

    for epoch in range(1, args.epochs + 1):
        # Adjust the learning rate for the SGD

        adjust_learning_rate(optimizer, epoch, args)

        # Commence training, timing the method and recording the time

        start = time.time()  # start timing the method

        train(args, model, device, train_loader, optimizer, ds_name, epoch)

        end = time.time()  # finish timing the method

        epoch_time = end - start

        # evaluation on natural examples (training and testing)
        print(
            "================================================================"
        )

        f.write("Epoch " + str(epoch) + "\n")

        print("Time for Training: " + str(epoch_time))

        f.write("Time for Training: " + str(epoch_time) + "\n")

        eval_clean(model, device, train_loader, "train", ds_name, f)

        eval_clean(model, device, test_loader, "test", ds_name, f)

        robust_eval(model, device, test_loader, ds_name, f)
        print(
            "================================================================"
        )

        # Save the model (if designated)

        if epoch % args.save_freq == 0:
            torch.save(
                model.state_dict(),
                os.path.join(
                    model_dir,
                    f"model-adt-{ds_name}-{mod_name}-epoch{epoch}.pt",
                ),
            )

            torch.save(
                optimizer.state_dict(),
                os.path.join(
                    model_dir,
                    f"opt-adt-{ds_name}-{mod_name}-epoch{epoch}.tar",
                ),
            )

    # Report the time taken to fully train the model
    end_tot = time.time()

    total_time = end_tot - start_tot

    print("Total training time: " + str(total_time))

    f.write("Total training time: " + str(total_time) + "\n")

    # Load the final trained model, make a deepcopy, and evaluate against FGSM and PGD

    if mod_name == "res18":
        trained_model = ResNet18(
            num_classes=100 if ds_name == "cifar100" else 10
        ).to(device)

    elif mod_name == "res34":
        trained_model = ResNet34(
            num_classes=100 if ds_name == "cifar100" else 10
        ).to(device)

    elif mod_name == "res50":
        trained_model = ResNet50(
            num_classes=100 if ds_name == "cifar100" else 10
        ).to(device)

    elif mod_name == "res101":
        trained_model = ResNet101(
            num_classes=100 if ds_name == "cifar100" else 10
        ).to(device)

    elif mod_name == "res152":
        trained_model = ResNet152(
            num_classes=100 if ds_name == "cifar100" else 10
        ).to(device)

    elif mod_name == "wideres34":
        trained_model = WideResNet(
            depth=34, num_classes=100 if ds_name == "cifar100" else 10
        ).to(device)

    else:
        raise NotImplementedError
    trained_model.load_state_dict(
        torch.load(
            f"data/model/model-adt-{ds_name}-{mod_name}-epoch{args.epochs}.pt"
        )
    )

    model_copy = copy.deepcopy(trained_model)

    robust_eval(model_copy, device, test_loader, ds_name, f)

    f.close()  # close the output file


if __name__ == "__main__":
    main_adt("cifar10")



# (c) 2024 LAiSR-SK

# This code is licensed under the MIT license (see LICENSE.md).

"""Train Adversarially Robust Models with Feature Scattering"""

import os
from collections import namedtuple
from os import PathLike

import torch
import torch.backends.cudnn as cudnn
import torchvision
import torchvision.transforms as transforms
from torch import optim
from tqdm import tqdm

from airtk.defense.feature_scatter.fs_attack_methods import (
    Attack_FeaScatter,
)
from airtk.model import WideResNet


class FeatureScatterTraining:
    def __init__(
        self,
        dataset: str,
        n_classes: int = 10,
        image_size: int = 32,
        epsilon: float = 8.0 / 512,
        lr: float = 0.1,
        num_steps: int = 1,
        step_size: int = 8.0 / 512,
        ls_factor: float = 0.5,
        epochs: int = 200,
        batch_size: int = 128,
        momentum: float = 0.9,
        save_epochs: int = 100,
        weight_decay: float = 2e-4,
        decay_epoch1: int = 60,
        decay_epoch2: int = 90,
        decay_rate: float = 0.1,
        log_step: int = 10,
        model_dir: PathLike = "data/model",
    ) -> None:
        ArgPrototype = namedtuple(
            "ArgPrototype",
            [
                "resume",
                "lr",
                "adv_mode",
                "model_dir",
                "init_model_pass",
                "max_epoch",
                "save_epochs",
                "decay_epoch1",
                "decay_epoch2",
                "decay_rate",
                "batch_size_train",
                "momentum",
                "weight_decay",
                "log_step",
                "num_classes",
                "image_size",
                "dataset",
            ],
        )

        if dataset == "cifar10":
            n_classes = 10

            image_size = 32

        elif dataset == "cifar100":
            n_classes = 100

            image_size = 32

        self._args = ArgPrototype(
            False,
            lr,
            "feature_scatter",
            model_dir,
            -1,
            epochs,
            save_epochs,
            decay_epoch1,
            decay_epoch2,
            decay_rate,
            batch_size,
            momentum,
            weight_decay,
            log_step,
            n_classes,
            image_size,
            dataset,
        )
        self._ds_name = dataset

        self._config = {
            "train": True,
            "epsilon": epsilon,
            "num_steps": num_steps,
            "step_size": step_size,
            "random_start": True,
            "ls_factor": ls_factor,
        }

    def __call__(self) -> None:
        main_featurescatter(self._ds_name, self._args, self._config)


# parser = argparse.ArgumentParser(description="Feature Scatterring Training")


# add type keyword to registries

# parser.register("type", "bool", str2bool)


# parser.add_argument(

# "--resume", "-r", action="store_true", help="resume from checkpoint"

# )

# parser.add_argument("--lr", default=0.1, type=float, help="learning rate")

# parser.add_argument(

# "--adv_mode",

# default="feature_scatter",

# type=str,

# help="adv_mode (feature_scatter)",

# )

# parser.add_argument(

# "--model_dir", default="\data\model", type=str, help="model path"

# )

# parser.add_argument(

# "--init_model_pass",

# default="-1",

# type=str,

# help="init model pass (-1: from scratch; K: checkpoint-K)",

# )

# parser.add_argument(

# "--max_epoch", default=200, type=int, help="max number of epochs"

# )

# parser.add_argument("--save_epochs", default=100, type=int, help="save period")

# parser.add_argument(

# "--decay_epoch1",

# default=60,

# type=int,

# help="learning rate decay epoch one",

# )

# parser.add_argument(

# "--decay_epoch2",

# default=90,

# type=int,

# help="learning rate decay point two",

# )

# parser.add_argument(

# "--decay_rate", default=0.1, type=float, help="learning rate decay rate"

# )

# parser.add_argument(

# "--batch_size_train", default=128, type=int, help="batch size for training"

# )

# parser.add_argument(

# "--momentum", default=0.9, type=float, help="momentum (1-tf.momentum)"

# )

# parser.add_argument(

# "--weight_decay", default=2e-4, type=float, help="weight decay"

# )

# parser.add_argument("--log_step", default=10, type=int, help="log_step")

#

# number of classes and image size will be updated below based on the dataset

# parser.add_argument("--num_classes", default=10, type=int, help="num classes")

# parser.add_argument("--image_size", default=32, type=int, help="image size")

# parser.add_argument(

# "--dataset", default="cifar10", type=str, help="dataset"

# )  # concat cascade


# args = parser.parse_args()

# device = "cuda" if torch.cuda.is_available() else "cpu"


def print_para(net):
    for name, param in net.named_parameters():
        if param.requires_grad:
            print(name)
            print(param.data)

        break


def zero_gradients(x):
    """Manually zeroes the gradients when pytorch is unable"""

    if x.grad is not None:
        x.grad.zero_()


def train_fun(epoch, net, trainloader, optimizer, args):
    print("\nEpoch: %d" % epoch)
    net.train()

    # update learning rate

    if epoch < args.decay_epoch1:
        lr = args.lr

    elif epoch < args.decay_epoch2:
        lr = args.lr * args.decay_rate

    else:
        lr = args.lr * args.decay_rate * args.decay_rate

    for param_group in optimizer.param_groups:
        param_group["lr"] = lr

    def get_acc(outputs, targets):
        _, predicted = outputs.max(1)

        total = targets.size(0)

        correct = predicted.eq(targets).sum().item()

        acc = 1.0 * correct / total
        return acc

    iterator = tqdm(trainloader, ncols=0, leave=False)

    device = "cuda" if torch.cuda.is_available() else "cpu"

    for batch_idx, (inputs, targets) in enumerate(iterator):
        inputs, targets = inputs.to(device), targets.to(device)

        adv_acc = 0

        optimizer.zero_grad()

        # forward
        outputs, loss_fs = net(inputs.detach(), targets)

        optimizer.zero_grad()
        loss = loss_fs.mean()

        loss.backward()

        optimizer.step()

        if batch_idx % args.log_step == 0:
            if adv_acc == 0:
                adv_acc = get_acc(outputs, targets)

            iterator.set_description(str(adv_acc))

            nat_outputs, _ = net(inputs, targets, attack=False)

            """print(

                "epoch %d, step %d, lr %.4f, duration %.2f, training nat acc %.2f, training adv acc %.2f, training adv loss %.4f"

                % (epoch, batch_idx, lr, duration, 100 * nat_acc,

                   100 * adv_acc, train_loss))"""

    if epoch % args.save_epochs == 0 or epoch >= args.max_epoch - 2:
        print("Saving..")

        f_path = os.path.join(args.model_dir, ("checkpoint-%s" % epoch))

        state = {
            "net": net.state_dict(),
            # 'optimizer': optimizer.state_dict()
        }

        if not os.path.isdir(args.model_dir):
            os.mkdir(args.model_dir)

        torch.save(state, f_path)

    if epoch >= 0:
        print("Saving latest @ epoch %s.." % (epoch))

        f_path = os.path.join(args.model_dir, "latest")

        state = {
            "net": net.state_dict(),
            "epoch": epoch,
            "optimizer": optimizer.state_dict(),
        }

        if not os.path.isdir(args.model_dir):
            os.mkdir(args.model_dir)

        torch.save(state, f_path)

        torch.save(
            net.state_dict(),
            f"data/model/feature-attack-epoch{epoch}.pt",
        )

        print("Epoch " + str(epoch) + " saved")


def main_featurescatter(ds_name, args, config_feature_scatter):
    start_epoch = 0

    transform_train = transforms.Compose(
        [
            transforms.RandomCrop(32, padding=4),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),  # [-1 1]
        ]
    )

    if args.dataset == "cifar10":
        trainset = torchvision.datasets.CIFAR10(
            root="./data", train=True, download=True, transform=transform_train
        )

    elif args.dataset == "cifar100":
        trainset = torchvision.datasets.CIFAR100(
            root="./data", train=True, download=True, transform=transform_train
        )

    else:
        raise NotImplementedError()

    trainloader = torch.utils.data.DataLoader(
        trainset, batch_size=args.batch_size_train, shuffle=True, num_workers=2
    )

    print("---wide resenet-----")

    device = "cuda" if torch.cuda.is_available() else "cpu"

    basic_net = WideResNet(
        depth=34, num_classes=args.num_classes, widen_factor=10
    ).to(device)

    if args.adv_mode.lower() == "feature_scatter":
        print("-----Feature Scatter mode -----")

        net = Attack_FeaScatter(basic_net, config_feature_scatter)

    else:
        print("-----OTHER_ALGO mode -----")

        raise NotImplementedError("Please implement this algorithm first!")

    if device == "cuda":
        net = torch.nn.DataParallel(net)

        cudnn.benchmark = True

    optimizer = optim.SGD(
        net.parameters(),
        lr=args.lr,
        momentum=args.momentum,
        weight_decay=args.weight_decay,
    )

    if args.resume and args.init_model_pass != "-1":
        # Load checkpoint.

        print("==> Resuming from checkpoint..")

        f_path_latest = os.path.join(args.model_dir, "latest")

        f_path = os.path.join(
            args.model_dir, ("checkpoint-%s" % args.init_model_pass)
        )

        if not os.path.isdir(args.model_dir):
            print("train from scratch: no checkpoint directory or file found")
        elif args.init_model_pass == "latest" and os.path.isfile(
            f_path_latest
        ):
            checkpoint = torch.load(f_path_latest)

            net.load_state_dict(checkpoint["net"])

            start_epoch = checkpoint["epoch"] + 1

            print("resuming from epoch %s in latest" % start_epoch)

        elif os.path.isfile(f_path):
            checkpoint = torch.load(f_path)

            net.load_state_dict(checkpoint["net"])

            start_epoch = checkpoint["epoch"] + 1

            print("resuming from epoch %s" % (start_epoch - 1))

        elif not os.path.isfile(f_path) or not os.path.isfile(f_path_latest):
            print("train from scratch: no checkpoint directory or file found")

    for epoch in range(1, args.max_epoch):
        train_fun(epoch, net, trainloader, optimizer, args)


if __name__ == "__main__":
    main_featurescatter("cifar10")



"""
Code is from https://github.com/BorealisAI/advertorch/blob/master/advertorch/attacks/spsa.py
"""

import torch
from torch.nn.modules.loss import _Loss

from ..attack import Attack


class MarginalLoss(_Loss):
    def forward(self, logits, targets):
        assert logits.shape[-1] >= 2
        top_logits, top_classes = torch.topk(logits, 2, dim=-1)
        target_logits = logits[torch.arange(logits.shape[0]), targets]
        max_nontarget_logits = torch.where(
            top_classes[..., 0] == targets,
            top_logits[..., 1],
            top_logits[..., 0],
        )

        loss = max_nontarget_logits - target_logits
        if self.reduction == "none":
            pass
        elif self.reduction == "sum":
            loss = loss.sum()
        elif self.reduction == "mean":
            loss = loss.mean()
        else:
            raise ValueError("unknown reduction: '%s'" % (self.recution,))

        return loss


class SPSA(Attack):
    r"""
    SPSA in the paper 'Adversarial Risk and the Dangers of Evaluating Against Weak Attacks'
    [https://arxiv.org/abs/1802.05666]

    Distance Measure : Linf

    Arguments:
        model (nn.Module): model to attack.
        eps (float): maximum perturbation. (Default: 8/255)
        delta (float): scaling parameter of SPSA. (Default: 0.01)
        lr (float): the learning rate of the `Adam` optimizer. (Default: 0.01)
        nb_iter (int): number of iterations of the attack. (Default: 1)
        nb_sample (int): number of samples for SPSA gradient approximation. (Default: 128)
        max_batch_size (int): maximum batch size to be evaluated at once. (Default: 64)

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.SPSA(model, eps=0.3)
        >>> adv_images = attack(images, labels)

    """

    def __init__(
        self,
        model,
        eps=0.3,
        delta=0.01,
        lr=0.01,
        nb_iter=1,
        nb_sample=128,
        max_batch_size=64,
    ):
        super().__init__("SPSA", model)
        self.eps = eps
        self.delta = delta
        self.lr = lr
        self.nb_iter = nb_iter
        self.nb_sample = nb_sample
        self.max_batch_size = max_batch_size
        self.loss_fn = MarginalLoss(reduction="none")
        self.supported_mode = ["default", "targeted"]

    def forward(self, images, labels):
        r"""
        Overridden.
        """

        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        adv_images = self.spsa_perturb(images, labels)
        return adv_images

    def loss(self, *args):
        if self.targeted:
            return self.loss_fn(*args)
        else:
            return -self.loss_fn(*args)

    def linf_clamp_(self, dx, x, eps):
        """Clamps perturbation `dx` to fit L_inf norm and image bounds.

        Limit the L_inf norm of `dx` to be <= `eps`, and the bounds of `x + dx`
        to be in `[clip_min, clip_max]`.

        Return: the clamped perturbation `dx`.
        """

        # dx_clamped = self.batch_clamp(eps, dx)
        dx_clamped = torch.clamp(dx, min=-eps, max=eps)
        # x_adv = self.clamp(x + dx_clamped, clip_min, clip_max)
        x_adv = torch.clamp(x + dx_clamped, min=0, max=1)
        # `dx` is changed *inplace* so the optimizer will keep
        # tracking it. the simplest mechanism for inplace was
        # adding the difference between the new value `x_adv - x`
        # and the old value `dx`.
        dx += x_adv - x - dx
        return dx

    def _get_batch_sizes(self, n, max_batch_size):
        batches = [max_batch_size for _ in range(n // max_batch_size)]
        if n % max_batch_size > 0:
            batches.append(n % max_batch_size)
        return batches

    @torch.no_grad()
    def spsa_grad(self, images, labels, delta, nb_sample, max_batch_size):
        """Uses SPSA method to apprixmate gradient w.r.t `x`.

        Use the SPSA method to approximate the gradient of `loss(predict(x), y)`
        with respect to `x`, based on the nonce `v`.

        Return the approximated gradient of `loss_fn(predict(x), y)` with respect to `x`.
        """

        grad = torch.zeros_like(images)
        images = torch.unsqueeze(images, 0)
        labels = torch.unsqueeze(labels, 0)

        def f(xvar, yvar):
            return self.loss(self.get_logits(xvar), yvar)

        images = images.expand(max_batch_size, *images.shape[1:]).contiguous()
        labels = labels.expand(max_batch_size, *labels.shape[1:]).contiguous()

        v = torch.empty_like(images[:, :1, ...])
        for batch_size in self._get_batch_sizes(nb_sample, max_batch_size):
            x_ = images[:batch_size]
            y_ = labels[:batch_size]
            vb = v[:batch_size]
            vb = vb.bernoulli_().mul_(2.0).sub_(1.0)
            v_ = vb.expand_as(x_).contiguous()
            x_shape = x_.shape
            x_ = x_.view(-1, *images.shape[2:])
            y_ = y_.view(-1, *labels.shape[2:])
            v_ = v_.view(-1, *v.shape[2:])
            df = f(x_ + delta * v_, y_) - f(x_ - delta * v_, y_)
            df = df.view(-1, *[1 for _ in v_.shape[1:]])
            grad_ = df / (2.0 * delta * v_)
            grad_ = grad_.view(x_shape)
            grad_ = grad_.sum(dim=0, keepdim=False)
            grad += grad_

        grad /= nb_sample
        return grad

    def spsa_perturb(self, x, y):
        dx = torch.zeros_like(x)
        dx.grad = torch.zeros_like(dx)
        optimizer = torch.optim.Adam([dx], lr=self.lr)
        for _ in range(self.nb_iter):
            optimizer.zero_grad()
            dx.grad = self.spsa_grad(
                x + dx, y, self.delta, self.nb_sample, self.max_batch_size
            )
            optimizer.step()
            dx = self.linf_clamp_(dx, x, self.eps)

        x_adv = x + dx
        return x_adv
